import numpy as np

from arpes.analysis import (boxcar_filter, curvature, d1_along_axis,
                            d2_along_axis, gaussian_filter)
from arpes.plotting.interactive_utils import BokehInteractiveTool
from arpes.utilities.funcutils import Debounce

__all__ = ['CurvatureTool']

class CurvatureTool(BokehInteractiveTool):
    # do not remove nans because they are used for masking in taking derivatives and smoothing
    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__()

        self.load_settings(**kwargs)
        self.app_main_size = self.settings.get('app_main_size', 400)

    def tool_handler(self, doc):
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models import widgets, Spacer
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.plotting import figure

        default_palette = self.default_palette

        x_coords, y_coords = self.arr.coords[self.arr.dims[1]], self.arr.coords[self.arr.dims[0]]
        self.app_context.update({
            'data': self.arr,
            'cached_data': {},
            'gamma_cached_data': {},
            'plots': {},
            'data_range': self.arr.T.range(),
            'figures': {},
            'widgets': {},
            'color_maps': {}})

        self.app_context['color_maps']['d2'] = LinearColorMapper(
            default_palette, low=np.min(self.arr.values),
            high=np.max(self.arr.values), nan_color='black')

        self.app_context['color_maps']['curvature'] = LinearColorMapper(
            default_palette, low=np.min(self.arr.values),
            high=np.max(self.arr.values), nan_color='black')

        self.app_context['color_maps']['raw'] = LinearColorMapper(
            default_palette, low=np.min(self.arr.values),
            high=np.max(self.arr.values), nan_color='black')

        plots, figures, data_range, cached_data, gamma_cached_data = (
            self.app_context['plots'], self.app_context['figures'], self.app_context['data_range'],
            self.app_context['cached_data'], self.app_context['gamma_cached_data'],
        )

        cached_data['raw'] = self.arr.values
        gamma_cached_data['raw'] = self.arr.values

        figure_kwargs = {
            'tools': ['reset', 'wheel_zoom'],
            'plot_width': self.app_main_size,
            'plot_height': self.app_main_size,
            'min_border': 10,
            'toolbar_location': 'left',
            'x_range': data_range['x'],
            'y_range': data_range['y'],
            'x_axis_location': 'below',
            'y_axis_location': 'right',
        }
        figures['d2'] = figure(title='d2 Spectrum', **figure_kwargs)

        figure_kwargs.update({
            'y_range': self.app_context['figures']['d2'].y_range,
            'x_range': self.app_context['figures']['d2'].x_range,
            'toolbar_location': None, 'y_axis_location': 'left',
        })

        figures['curvature'] = figure(title='Curvature', **figure_kwargs)
        figures['raw'] = figure(title='Raw Image', **figure_kwargs)

        figures['curvature'].yaxis.major_label_text_font_size = '0pt'

        # TODO add support for color mapper
        plots['d2'] = figures['d2'].image(
            [self.arr.values], x=data_range['x'][0], y=data_range['y'][0],
            dw=data_range['x'][1] - data_range['x'][0], dh=data_range['y'][1] - data_range['y'][0],
            color_mapper=self.app_context['color_maps']['d2']
        )
        plots['curvature'] = figures['curvature'].image(
            [self.arr.values], x=data_range['x'][0], y=data_range['y'][0],
            dw=data_range['x'][1] - data_range['x'][0], dh=data_range['y'][1] - data_range['y'][0],
            color_mapper=self.app_context['color_maps']['curvature']
        )
        plots['raw'] = figures['raw'].image(
            [self.arr.values], x=data_range['x'][0], y=data_range['y'][0],
            dw=data_range['x'][1] - data_range['x'][0], dh=data_range['y'][1] - data_range['y'][0],
            color_mapper=self.app_context['color_maps']['raw']
        )

        smoothing_sliders_by_name = {}
        smoothing_sliders = []  # need one for each axis
        axis_resolution = self.arr.T.stride(generic_dim_names=False)
        for dim in self.arr.dims:
            coords = self.arr.coords[dim]
            resolution = float(axis_resolution[dim])
            high_resolution = len(coords) / 3 * resolution
            low_resolution = resolution

            # could make this axis dependent for more reasonable defaults
            default = 15 * resolution

            if default > high_resolution:
                default = (high_resolution + low_resolution) / 2

            new_slider = widgets.Slider(
                title='{} Window'.format(dim), start=low_resolution, end=high_resolution,
                step=resolution, value=default)
            smoothing_sliders.append(new_slider)
            smoothing_sliders_by_name[dim] = new_slider

        n_smoothing_steps_slider = widgets.Slider(
            title="Smoothing Steps", start=0, end=5, step=1, value=2)
        beta_slider = widgets.Slider(
            title="β", start=-8, end=8, step=1, value=0)
        direction_select = widgets.Select(
            options=list(self.arr.dims),
            value='eV' if 'eV' in self.arr.dims else self.arr.dims[0],  # preference to energy,
            title='Derivative Direction'
        )
        interleave_smoothing_toggle = widgets.Toggle(
            label='Interleave smoothing with d/dx', active=True,
            button_type='primary')
        clamp_spectrum_toggle = widgets.Toggle(
            label='Clamp positive values to 0', active=True,
            button_type='primary')
        filter_select = widgets.Select(
            options=['Gaussian', 'Boxcar'],
            value='Boxcar',
            title='Type of Filter'
        )

        color_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Clip')
        gamma_slider = widgets.Slider(
            start=0.1, end=4, value=1, step=0.1, title='Gamma')

        # don't need any cacheing here for now, might if this ends up being too slow
        def smoothing_fn(n_passes):
            if n_passes == 0:
                return lambda x: x

            filter_factory = {
                'Gaussian': gaussian_filter,
                'Boxcar': boxcar_filter,
            }.get(filter_select.value, boxcar_filter)

            filter_size = {d: smoothing_sliders_by_name[d].value for d in self.arr.dims}
            return filter_factory(filter_size, n_passes)

        @Debounce(0.25)
        def force_update():
            n_smoothing_steps = n_smoothing_steps_slider.value
            d2_data = self.arr
            if interleave_smoothing_toggle.active:
                f = smoothing_fn(n_smoothing_steps // 2)
                d2_data = d1_along_axis(f(d2_data), direction_select.value)
                f = smoothing_fn(n_smoothing_steps - (n_smoothing_steps // 2))
                d2_data = d1_along_axis(f(d2_data), direction_select.value)

            else:
                f = smoothing_fn(n_smoothing_steps)
                d2_data = d2_along_axis(f(self.arr), direction_select.value)

            d2_data.values[
                d2_data.values != d2_data.values] = 0  # remove NaN values until Bokeh fixes NaNs over the wire
            if clamp_spectrum_toggle.active:
                d2_data.values = -d2_data.values
                d2_data.values[d2_data.values < 0] = 0
            cached_data['d2'] = d2_data.values
            gamma_cached_data['d2'] = d2_data.values ** gamma_slider.value
            plots['d2'].data_source.data = {
                'image': [gamma_cached_data['d2']]
            }

            curv_smoothing_fn = smoothing_fn(n_smoothing_steps)
            smoothed_curvature_data = curv_smoothing_fn(self.arr)
            curvature_data = curvature(smoothed_curvature_data, self.arr.dims, beta=beta_slider.value)
            curvature_data.values[curvature_data.values != curvature_data.values] = 0
            if clamp_spectrum_toggle.active:
                curvature_data.values = -curvature_data.values
                curvature_data.values[curvature_data.values < 0] = 0

            cached_data['curvature'] = curvature_data.values
            gamma_cached_data['curvature'] = curvature_data.values ** gamma_slider.value
            plots['curvature'].data_source.data = {
                'image': [gamma_cached_data['curvature']]
            }
            update_color_slider(color_slider.value)

        # TODO better integrate these, they can share code with the above if we are more careful.
        def take_d2(d2_data):
            n_smoothing_steps = n_smoothing_steps_slider.value
            if interleave_smoothing_toggle.active:
                f = smoothing_fn(n_smoothing_steps // 2)
                d2_data = d1_along_axis(f(d2_data), direction_select.value)
                f = smoothing_fn(n_smoothing_steps - (n_smoothing_steps // 2))
                d2_data = d1_along_axis(f(d2_data), direction_select.value)

            else:
                f = smoothing_fn(n_smoothing_steps)
                d2_data = d2_along_axis(f(self.arr), direction_select.value)

            d2_data.values[
                d2_data.values != d2_data.values] = 0  # remove NaN values until Bokeh fixes NaNs over the wire
            if clamp_spectrum_toggle.active:
                d2_data.values = -d2_data.values
                d2_data.values[d2_data.values < 0] = 0

            return d2_data

        def take_curvature(curvature_data, curve_dims):
            curv_smoothing_fn = smoothing_fn(n_smoothing_steps_slider.value)
            smoothed_curvature_data = curv_smoothing_fn(curvature_data)
            curvature_data = curvature(smoothed_curvature_data, curve_dims, beta=beta_slider.value)
            curvature_data.values[curvature_data.values != curvature_data.values] = 0
            if clamp_spectrum_toggle.active:
                curvature_data.values = -curvature_data.values
                curvature_data.values[curvature_data.values < 0] = 0

            return curvature_data

        # These functions will always be linked to the current context of the curvature tool.
        self.app_context['d2_fn'] = take_d2
        self.app_context['curvature_fn'] = take_curvature

        def force_update_change_wrapper(attr, old, new):
            if old != new:
                force_update()

        def force_update_click_wrapper(event):
            force_update()

        @Debounce(0.1)
        def update_color_slider(new):
            def update_plot(name, data):
                low, high = np.min(data), np.max(data)
                dynamic_range = high - low
                self.app_context['color_maps'][name].update(
                    low=low + new[0] / 100 * dynamic_range, high=low + new[1] / 100 * dynamic_range)

            update_plot('d2', gamma_cached_data['d2'])
            update_plot('curvature', gamma_cached_data['curvature'])
            update_plot('raw', gamma_cached_data['raw'])

        @Debounce(0.1)
        def update_gamma_slider(new):
            gamma_cached_data['d2'] = cached_data['d2'] ** new
            gamma_cached_data['curvature'] = cached_data['curvature'] ** new
            gamma_cached_data['raw'] = cached_data['raw'] ** new
            update_color_slider(color_slider.value)

        def update_color_handler(attr, old, new):
            update_color_slider(new)

        def update_gamma_handler(attr, old, new):
            update_gamma_slider(new)

        layout = column(
            row(
                column(self.app_context['figures']['d2'],
                       interleave_smoothing_toggle,
                       direction_select),
                column(self.app_context['figures']['curvature'],
                       beta_slider, clamp_spectrum_toggle),
                column(self.app_context['figures']['raw'],
                       color_slider, gamma_slider)
            ),
            widgetbox(
                filter_select,
                *smoothing_sliders,
                n_smoothing_steps_slider,
            ),
            Spacer(height=100),
        )

        # Attach event handlers
        for w in (n_smoothing_steps_slider, beta_slider, direction_select,
                  *smoothing_sliders, filter_select):
            w.on_change('value', force_update_change_wrapper)

        interleave_smoothing_toggle.on_click(force_update_click_wrapper)
        clamp_spectrum_toggle.on_click(force_update_click_wrapper)

        color_slider.on_change('value', update_color_handler)
        gamma_slider.on_change('value', update_gamma_handler)

        force_update()

        doc.add_root(layout)
        doc.title = 'Curvature Tool'



# pylint: disable=import-error

import pyqtgraph as pg

__all__ = ('DataArrayImageView',)

class DataArrayImageView(pg.ImageView):
    """
    ImageView that transparently handles xarray data, including setting axis and coordinate information.

    This makes it easier to build interactive applications around realistic scientific datasets.
    """
    def __init__(self, root, *args, **kwargs):
        super().__init__(view=pg.PlotItem(), *args, **kwargs)

        self.view.invertY(False)
        self.root = root

    def setImage(self, img, *args, **kwargs):
        """
        Accepts an xarray.DataArray instead of a numpy array
        :param img:
        :param args:
        :param kwargs:
        :return:
        """

        super().setImage(img.values, *args, **kwargs)

    def recompute(self):
        pass



from functools import wraps

from arpes.provenance import provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from typing import List

__all__ = ('decomposition_along', 'pca_along', 'ica_along', 'factor_analysis_along',)


def decomposition_along(data: DataType, axes: List[str], decomposition_cls, correlation=False, **kwargs):
    """
    Performs a change of basis of your data according to `sklearn` decomposition classes. This allows
    for robust and simple PCA, ICA, factor analysis, and other decompositions of your data even when it
    is very high dimensional.

    Generally speaking, PCA and similar techniques work when data is 2D, i.e. a sequence of 1D observations.
    We can make the same techniques work by unravelling a ND dataset into 1D (i.e. np.ndarray.ravel()) and
    unravelling a KD set of observations into a 1D set of observations. This is basically grouping axes. As
    an example, if you had a 4D dataset which consisted of 2D-scanning valence band ARPES, then the dimensions
    on our dataset would be "[x,y,eV,phi]". We can group these into [spatial=(x, y), spectral=(eV, phi)] and
    perform PCA or another analysis of the spectral features over different spatial observations.

    If our data was called `f`, this can be accomplished with:

    ```
    transformed, decomp = decomposition_analysis(f.stack(spectral=['eV', 'phi']), ['x', 'y'], PCA)
    transformed.dims # -> [X, Y, components]
    ```

    The results of `decomposition_along` can be explored with `arpes.widgets.pca_explorer`, regardless of
    the decomposition class.

    :param data: Input data, can be N-dimensional but should only include one "spectral" axis.
    :param axes: Several axes to be treated as a single axis labeling the list of observations.
    :param decomposition_cls: A sklearn.decomposition class (such as PCA or ICA) to be used
                              to perform the decomposition.
    :param correlation: Controls whether StandardScaler() is used as the first stage of the data ingestion
                        pipeline for sklearn.
    :param kwargs:
    :return:
    """
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler


    if len(axes) > 1:
        flattened_data = normalize_to_spectrum(data).stack(fit_axis=axes)
        stacked = True
    else:
        flattened_data = normalize_to_spectrum(data).S.transpose_to_back(axes[0])
        stacked = False

    if len(flattened_data.dims) != 2:
        raise ValueError('Inappropriate number of dimensions after flattening: [{}]'.format(
            flattened_data.dims))

    if correlation:
        pipeline = make_pipeline(StandardScaler(), decomposition_cls(**kwargs))
    else:
        pipeline = make_pipeline(decomposition_cls(**kwargs))

    pipeline.fit(flattened_data.values.T)

    decomp = pipeline.steps[-1][1]

    transform = decomp.transform(flattened_data.values.T)

    into = flattened_data.copy(deep=True)
    into_first = into.dims[0]
    into = into.isel(**dict([[into_first, slice(0, transform.shape[1])]]))
    into = into.rename(dict([[into_first, 'components']]))

    into.values = transform.T

    if stacked:
        into = into.unstack('fit_axis')

    provenance(into, data, {
        'what': 'sklearn decomposition',
        'by': 'decomposition_along',
        'axes': axes,
        'correlation': False,
        'decomposition_cls': decomposition_cls.__name__,
    })

    return into, decomp


@wraps(decomposition_along)
def pca_along(*args, **kwargs):
    from sklearn.decomposition import PCA
    return decomposition_along(*args, **kwargs, decomposition_cls=PCA)


@wraps(decomposition_along)
def factor_analysis_along(*args, **kwargs):
    from sklearn.decomposition import FactorAnalysis
    return decomposition_along(*args, **kwargs, decomposition_cls=FactorAnalysis)


@wraps(decomposition_along)
def ica_along(*args, **kwargs):
    from sklearn.decomposition import FastICA
    return decomposition_along(*args, **kwargs, decomposition_cls=FastICA)



# pylint: disable=import-error

from collections import namedtuple

from PyQt5 import QtCore, QtWidgets

__all__ = ('PRETTY_KEYS', 'pretty_key_event', 'KeyBinding', 'layout', 'hlayout', 'vlayout', 'tabs', 'label')

KeyBinding = namedtuple('KeyBinding', ('label', 'chord', 'handler'))

PRETTY_KEYS = {}
for key, value in vars(QtCore.Qt).items():
    if isinstance(value, QtCore.Qt.Key):
        PRETTY_KEYS[value] = key.partition('_')[2]


def pretty_key_event(event):
    """
    Key Event -> List[str] in order to be able to prettily print keys
    :param event:
    :return:
    """
    key_sequence = []

    key_name = PRETTY_KEYS.get(event.key(), event.text())
    if key_name not in key_sequence:
        key_sequence.append(key_name)

    return key_sequence


def tabs(*children):
    widget = QtWidgets.QTabWidget()
    for name, child in children:
        widget.addTab(child, name)

    return widget


def layout(*children, layout_cls=None, widget=None):
    if layout_cls is None:
        layout_cls = QtWidgets.QGridLayout

    if widget is None:
        widget = QtWidgets.QWidget()

    internal_layout = layout_cls()

    for child in children:
        internal_layout.addWidget(child)


    widget.setLayout(internal_layout)

    return widget


def hlayout(*args, **kwargs):
    return layout(*args, **kwargs, layout_cls=QtWidgets.QHBoxLayout)


def vlayout(*args, **kwargs):
    return layout(*args, **kwargs, layout_cls=QtWidgets.QVBoxLayout)


def label(text, **kwargs):
    return QtWidgets.QLabel(text, **kwargs)



import copy
import os.path
import warnings

import h5py
import numpy as np

import arpes.config
import xarray as xr
from arpes.endstations import EndstationBase
from arpes.provenance import provenance_from_file

__all__ = ('SToFDLDEndstation',)

class SToFDLDEndstation(EndstationBase):
    PRINCIPAL_NAME = 'ALG-SToF-DLD'

    def load(self, scan_desc: dict=None, **kwargs):
        """
        Imports a FITS file that contains all of the information from a run of Ping
        and Anton's delay line detector ARToF

        :param scan_desc: Dictionary with extra information to attach to the xarray.Dataset, must contain the location
        of the file
        :return: xarray.Dataset
        """

        if scan_desc is None:
            warnings.warn('Attempting to make due without user associated metadata for the file')
            raise TypeError('Expected a dictionary of metadata with the location of the file')

        metadata = copy.deepcopy(scan_desc)

        data_loc = metadata['file']
        data_loc = data_loc if data_loc.startswith('/') else os.path.join(arpes.config.DATA_PATH, data_loc)

        f = h5py.File(data_loc, 'r')

        dataset_contents = dict()
        raw_data = f['/PRIMARY/DATA'][:]
        raw_data = raw_data[:, ::-1]  # Reverse the timing axis
        dataset_contents['raw'] = xr.DataArray(
            raw_data,
            coords={'x_pixels': np.linspace(0, 511, 512),
                    't_pixels': np.linspace(0, 511, 512)},
            dims=('x_pixels', 't_pixels'),
            attrs=f['/PRIMARY'].attrs.items(),
        )

        provenance_from_file(dataset_contents['raw'], data_loc, {
            'what': 'Loaded Anton and Ping DLD dataset from HDF5.',
            'by': 'load_DLD',
        })

        return xr.Dataset(
            dataset_contents,
            attrs=metadata
        )



import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import xarray as xr
from arpes.analysis import rebin
from arpes.plotting.tof import scatter_with_std
from arpes.plotting.utils import (colorbarmaps_for_axis, generic_colorbarmap_for_data, fancy_labels,
                                  path_for_plot, label_for_dim)
from arpes.provenance import save_plot_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('stack_dispersion_plot', 'flat_stack_plot',)


@save_plot_provenance
def offset_scatter_plot(data: DataType, name_to_plot=None, stack_axis=None, fermi_level=True, cbarmap=None, ax=None,
                        out=None, scale_coordinate=0.5, ylim=None, aux_errorbars=True, **kwargs):
    assert isinstance(data, xr.Dataset)

    if name_to_plot is None:
        var_names = [k for k in data.data_vars.keys() if '_std' not in k]
        assert len(var_names) == 1
        name_to_plot = var_names[0]
        assert (name_to_plot + '_std') in data.data_vars.keys()

    if len(data.data_vars[name_to_plot].dims) != 2:
        raise ValueError('In order to produce a stack plot, data must be image-like.'
                         'Passed data included dimensions: {}'.format(data.data_vars[name_to_plot].dims))

    fig = None
    inset_ax = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (11, 5,)))

    if inset_ax is None:
        inset_ax = inset_axes(ax, width='40%', height='5%', loc='upper left')

    if stack_axis is None:
        stack_axis = data.data_vars[name_to_plot].dims[0]

    skip_colorbar = True
    if cbarmap is None:
        skip_colorbar = False
        try:
            cbarmap = colorbarmaps_for_axis[stack_axis]
        except:
            cbarmap = generic_colorbarmap_for_data(data.coords[stack_axis], ax=inset_ax, ticks=kwargs.get('ticks'))

    cbar, cmap = cbarmap

    if not isinstance(cmap, matplotlib.colors.Colormap):
        # do our best
        try:
            cmap = cmap()
        except:
            # might still be fine
            pass

    # should be exactly two
    other_dim = [d for d in data.dims if d != stack_axis][0]
    other_coord = data.coords[other_dim]

    if 'eV' in data.dims and 'eV' != stack_axis and fermi_level:
        ax.axhline(0, linestyle='--', color='red')
        ax.fill_betweenx([-1e6, 1e6], 0, 0.2, color='black', alpha=0.07)
        ax.set_ylim(ylim)

    # real plotting here
    for i, (coord, value) in enumerate(data.T.iterate_axis(stack_axis)):
        delta = data.T.stride(generic_dim_names=False)[other_dim]
        data_for = value.copy(deep=True)
        data_for.coords[other_dim] = data_for.coords[other_dim].copy(deep=True)
        data_for.coords[other_dim].values = data_for.coords[other_dim].values.copy()
        data_for.coords[other_dim].values -= i * delta * scale_coordinate / 10

        scatter_with_std(data_for, name_to_plot, ax=ax, color=cmap(coord[stack_axis]))

        if aux_errorbars:
            assert ylim is not None
            data_for = data_for.copy(deep=True)
            flattened = data_for.data_vars[name_to_plot].copy(deep=True)
            flattened.values = ylim[0] * np.ones(flattened.values.shape)
            data_for = data_for.assign(**{name_to_plot: flattened})
            scatter_with_std(data_for, name_to_plot, ax=ax, color=cmap(coord[stack_axis]), fmt='none')


    ax.set_xlabel(other_dim)
    ax.set_ylabel(name_to_plot)
    fancy_labels(ax)

    try:
        if inset_ax and not skip_colorbar:
            inset_ax.set_xlabel(stack_axis, fontsize=16)

            fancy_labels(inset_ax)
            cbar(ax=inset_ax, **kwargs)
    except TypeError:
        # colorbar already rendered
        pass

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax


@save_plot_provenance
def flat_stack_plot(data: DataType, stack_axis=None, fermi_level=True, cbarmap=None, ax=None,
                    mode='line', title=None, out=None, transpose=False, **kwargs):
    data = normalize_to_spectrum(data)
    if len(data.dims) != 2:
        raise ValueError('In order to produce a stack plot, data must be image-like.'
                         'Passed data included dimensions: {}'.format(data.dims))

    fig = None
    inset_ax = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (7, 5,)))
        inset_ax = inset_axes(ax, width='40%', height='5%', loc=1)

    if stack_axis is None:
        stack_axis = data.dims[0]

    skip_colorbar = True
    if cbarmap is None:
        skip_colorbar = False
        try:
            cbarmap = colorbarmaps_for_axis[stack_axis]
        except KeyError:
            cbarmap = generic_colorbarmap_for_data(data.coords[stack_axis], ax=inset_ax, ticks=kwargs.get('ticks'))

    cbar, cmap = cbarmap

    # should be exactly two
    other_dim = [d for d in data.dims if d != stack_axis][0]
    other_coord = data.coords[other_dim]

    if not isinstance(cmap, matplotlib.colors.Colormap):
        # do our best
        try:
            cmap = cmap()
        except:
            # might still be fine
            pass

    if 'eV' in data.dims and 'eV' != stack_axis and fermi_level:
        if transpose:
            ax.axhline(0, color='red', alpha=0.8, linestyle='--', linewidth=1)
        else:
            ax.axvline(0, color='red', alpha=0.8, linestyle='--', linewidth=1)

    # meat of the plotting
    for coord_dict, marginal in list(data.T.iterate_axis(stack_axis)):
        if transpose:
            if mode == 'line':
                ax.plot(marginal.values, marginal.coords[marginal.dims[0]].values, color=cmap(coord_dict[stack_axis]), **kwargs)
            else:
                assert mode == 'scatter'
                raise NotImplementedError()
        else:
            if mode == 'line':
                marginal.plot(ax=ax, color=cmap(coord_dict[stack_axis]), **kwargs)
            else:
                assert mode == 'scatter'
                ax.scatter(*marginal.T.to_arrays(), color=cmap(coord_dict[stack_axis]))
                ax.set_xlabel(marginal.dims[0])

    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel('Spectrum Intensity (arb).')
    ax.set_title(title, fontsize=14)
    ax.set_xlim([other_coord.min().item(), other_coord.max().item()])

    try:
        if inset_ax is not None and not skip_colorbar:
            inset_ax.set_xlabel(stack_axis, fontsize=16)
            fancy_labels(inset_ax)

            cbar(ax=inset_ax, **kwargs)
    except TypeError:
        # already rendered
        pass

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax


@save_plot_provenance
def stack_dispersion_plot(data: DataType, stack_axis=None, ax=None, title=None, out=None,
                          max_stacks=100, transpose=False,
                          use_constant_correction=False, correction_side=None,
                          color=None, c=None,
                          label=None,
                          shift=0,
                          no_scatter=False,
                          negate=False, s=1, scale_factor=None, linewidth=1, palette=None, zero_offset=False, uniform = False, **kwargs):
    data = normalize_to_spectrum(data)

    if stack_axis is None:
        stack_axis = data.dims[0]

    other_axes = list(data.dims)
    other_axes.remove(stack_axis)
    other_axis = other_axes[0]

    stack_coord = data.coords[stack_axis]
    if len(stack_coord.values) > max_stacks:
        data = rebin(data, reduction=dict([[
            stack_axis, int(np.ceil(len(stack_coord.values) / max_stacks))]
        ]))

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    if title is None:
        title = '{} Stack'.format(data.S.label.replace('_', ' '))

    max_over_stacks = np.max(data.values)

    cvalues = data.coords[other_axis].values
    if scale_factor is None:
        maximum_deviation = -np.inf

        for _, marginal in data.T.iterate_axis(stack_axis):
            marginal_values = -marginal.values if negate else marginal.values
            marginal_offset, right_marginal_offset = marginal_values[0], marginal_values[-1]

            if use_constant_correction:
                true_ys = (marginal_values - marginal_offset)
            elif zero_offset:
                true_ys = marginal_values 
            else:
                true_ys = (marginal_values - np.linspace(marginal_offset, right_marginal_offset, len(marginal_values)))

            maximum_deviation = np.max([maximum_deviation] + list(np.abs(true_ys)))

        scale_factor = 0.02 * (np.max(cvalues) - np.min(cvalues)) / maximum_deviation

    iteration_order = -1 # might need to fiddle with this in certain cases
    lim = [-np.inf, np.inf]
    labeled = False
    for i, (coord_dict, marginal) in enumerate(list(data.T.iterate_axis(stack_axis))[::iteration_order]):
        coord_value = coord_dict[stack_axis]

        xs = cvalues
        marginal_values = -marginal.values if negate else marginal.values
        marginal_offset, right_marginal_offset = marginal_values[0], marginal_values[-1]

        if use_constant_correction:
            offset = right_marginal_offset if correction_side == 'right' else marginal_offset
            true_ys = (marginal_values - offset) / max_over_stacks
            ys = scale_factor * true_ys + coord_value
        elif zero_offset:
            true_ys = marginal_values / max_over_stacks
            ys = scale_factor * true_ys + coord_value
        elif uniform:
            true_ys = marginal_values / max_over_stacks
            ys = scale_factor * true_ys + i            
        else:
            true_ys = (marginal_values - np.linspace(marginal_offset, right_marginal_offset, len(marginal_values))) \
                      / max_over_stacks
            ys = scale_factor * true_ys + coord_value

        raw_colors = color or c or 'black'

        if palette:
            if isinstance(palette, str):
                palette = cm.get_cmap(palette)
            raw_colors = palette(np.abs(true_ys / max_over_stacks))

        if transpose:
            xs, ys = ys, xs

        xs = xs - i * shift

        lim = [max(lim[0], np.min(xs)), min(lim[1], np.max(xs))]

        label_for = '_nolegend_'
        if not labeled:
            labeled = True
            label_for = label

        color_for_plot = raw_colors
        if callable(color_for_plot):
            color_for_plot = color_for_plot(coord_value)

        if isinstance(raw_colors, (str, tuple)) or no_scatter:
            ax.plot(xs, ys, linewidth=linewidth, color=color_for_plot, label=label_for, **kwargs)
        else:
            ax.scatter(xs, ys, color=color_for_plot, s=s, label=label_for, **kwargs)

    x_label = other_axis
    y_label = stack_axis

    if transpose:
        x_label, y_label = y_label, x_label

    ax.set_xlabel(label_for_dim(data, x_label))
    ax.set_ylabel(label_for_dim(data, y_label))

    if transpose:
        ax.set_ylim(lim)
    else:
        ax.set_xlim(lim)

    ax.set_title(title)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax


@save_plot_provenance
def overlapped_stack_dispersion_plot(data: DataType, stack_axis=None, ax=None, title=None, out=None,
                                     max_stacks=100, use_constant_correction=False, transpose=False,
                                     negate=False, s=1, scale_factor=None, linewidth=1, palette=None, **kwargs):
    data = normalize_to_spectrum(data)

    if stack_axis is None:
        stack_axis = data.dims[0]

    other_axes = list(data.dims)
    other_axes.remove(stack_axis)
    other_axis = other_axes[0]

    stack_coord = data.coords[stack_axis]
    if len(stack_coord.values) > max_stacks:
        data = rebin(data, reduction=dict([[
            stack_axis, int(np.ceil(len(stack_coord.values) / max_stacks))]
        ]))

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    if title is None:
        title = '{} Stack'.format(data.S.label.replace('_', ' '))

    max_over_stacks = np.max(data.values)

    cvalues = data.coords[other_axis].values
    if scale_factor is None:
        maximum_deviation = -np.inf

        for _, marginal in data.T.iterate_axis(stack_axis):
            marginal_values = -marginal.values if negate else marginal.values
            marginal_offset, right_marginal_offset = marginal_values[0], marginal_values[-1]

            if use_constant_correction:
                true_ys = (marginal_values - marginal_offset)
            else:
                true_ys = (marginal_values - np.linspace(marginal_offset, right_marginal_offset, len(marginal_values)))

            maximum_deviation = np.max([maximum_deviation] + list(np.abs(true_ys)))

        scale_factor = 0.02 * (np.max(cvalues) - np.min(cvalues)) / maximum_deviation

    iteration_order = -1 # might need to fiddle with this in certain cases
    for coord_dict, marginal in list(data.T.iterate_axis(stack_axis))[::iteration_order]:
        coord_value = coord_dict[stack_axis]

        xs = cvalues
        marginal_values = -marginal.values if negate else marginal.values
        marginal_offset, right_marginal_offset = marginal_values[0], marginal_values[-1]

        if use_constant_correction:
            true_ys = (marginal_values - marginal_offset) / max_over_stacks
            ys = scale_factor * true_ys + coord_value
        else:
            true_ys = (marginal_values - np.linspace(marginal_offset, right_marginal_offset, len(marginal_values))) \
                      / max_over_stacks
            ys = scale_factor * true_ys + coord_value

        raw_colors = 'black'
        if palette:
            if isinstance(palette, str):
                palette = cm.get_cmap(palette)
            raw_colors = palette(np.abs(true_ys / max_over_stacks))

        if transpose:
            xs, ys = ys, xs

        if isinstance(raw_colors, str):
            plt.plot(xs, ys, linewidth=linewidth, color=raw_colors, **kwargs)
        else:
            plt.scatter(xs, ys, color=raw_colors, s=s, **kwargs)

    x_label = other_axis
    y_label = stack_axis

    if transpose:
        x_label, y_label = y_label, x_label

    ax.set_xlabel(label_for_dim(data, x_label))
    ax.set_ylabel(label_for_dim(data, y_label))

    ax.set_title(title)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()

    return fig, ax



"""
Small utilities for estimating quantities of practical interest when using
a laser for photoemission.
"""

from arpes.config import ureg

__all__ = ('electrons_per_pulse',)

mira_frequency = (54.3 / ureg.microsecond)


def electrons_per_pulse(photocurrent, division_ratio=None, pulse_rate=None):
    assert(division_ratio is None or pulse_rate is None)

    if division_ratio is not None:
        pulse_rate = mira_frequency / division_ratio

    eles_per_attocoulomb = 6.2415091
    atto_coulombs = (photocurrent / pulse_rate).to('attocoulomb')

    return (atto_coulombs * eles_per_attocoulomb).magnitude



"""
Provides data provenance for PyARPES. Most analysis routines built
into PyARPES support provenance. Of course, Python is a dynamic language and nothing can be
done to prevent the experimenter from circumventing the provenance scheme.

All the same, between analysis notebooks and the data provenenace provided by PyARPES,
we provide an environment with much higher standard for reproducible analysis than many
other current analysis environments.

This provenenace record is automatically exported when using the built in
plotting utilities. Additionally, passing `used_data` to the PyARPES `savefig`
wrapper allows saving provenance information even for bespoke plots created in
a Jupyter cell.

PyARPES also makes it easy to opt into data provenance for new analysis
functions by providing convenient decorators. These decorators inspect data passed at runtime
to look for and update provenance entries on arguments and return values.
"""

import datetime
import functools
import json
import os.path
import typing
import uuid
import warnings

from typing import Any

import xarray as xr

import arpes.config
from arpes import VERSION
from arpes.typing import xr_types, DataType


def attach_id(data: xr.DataArray) -> None:
    """
    Ensures that an ID is attached to a piece of data, if it does not already exist.
    IDs are generated at the time of identification in an analysis notebook. Sometimes a piece of
    data is created from nothing, and we might need to generate one for it on the spot.
    :param data:
    :return:
    """
    if 'id' not in data.attrs:
        data.attrs['id'] = str(uuid.uuid1())


def provenance_from_file(child_arr: typing.Union[xr.DataArray, xr.Dataset], file, record):
    """
    Builds a provenance entry for a dataset corresponding to loading data from a file. This is used
    by data loaders at the start of an analysis.
    :param child_arr:
    :param file:
    :param record:
    :return:
    """

    if 'id' not in child_arr.attrs:
        attach_id(child_arr)

    child_arr.attrs['provenance'] = {
        'record': record,
        'file': file,
        'parents_provenance': 'filesystem',
        'time': datetime.datetime.now().isoformat(),
        'version': VERSION,
    }


def update_provenance(what, record_args=None, keep_parent_ref=False):
    """
    Provides a decorator that promotes a function to one that records data provenance.

    :param what: Description of what transpired, to put into the record.
    :param record_args: Unused presently, will allow recording args into record.
    :param keep_parent_ref: Whether to keep a pointer to the parents in the hierarchy or not.
    :return: decorator
    """
    def update_provenance_decorator(fn):
        @functools.wraps(fn)
        def func_wrapper(*args: Any, **kwargs: Any) -> xr.DataArray:

            arg_parents = [v for v in args if isinstance(v, xr_types) and 'id' in v.attrs]
            kwarg_parents = {k: v for k, v in kwargs.items()
                             if isinstance(v, xr_types) and 'id' in v.attrs}
            all_parents = arg_parents + list(kwarg_parents.values())
            result = fn(*args, **kwargs)

            # we do not want to record provenance or change the id if ``f`` opted not to do anything
            # to its input. This reduces the burden on client code by allowing them to return the input
            # without changing the 'id' attr
            result_not_identity = not any(p is result for p in all_parents)

            if isinstance(result, xr_types) and result_not_identity:
                if 'id' in result.attrs:
                    del result.attrs['id']

                provenance_fn = provenance
                if len(all_parents) > 1:
                    provenance_fn = provenance_multiple_parents

                if all_parents:
                    provenance_fn(result, all_parents, {
                        'what': what,
                        'by': fn.__name__,
                        'time': datetime.datetime.now().isoformat(),
                        'version': VERSION,
                    }, keep_parent_ref)

            return result
        return func_wrapper
    return update_provenance_decorator


def save_plot_provenance(plot_fn):
    """
    A decorator that automates saving the provenance information for a particular plot.
    A plotting function creates an image or movie resource at some location on the
    filesystem.

    In order to hook into this decorator appropriately, because there is no way that I know
    of of temporarily overriding the behavior of the open builtin in order to monitor
    for a write.

    :param plot_fn: A plotting function to decorate
    :return:
    """
    @functools.wraps(plot_fn)
    def func_wrapper(*args, **kwargs):
        path = plot_fn(*args, **kwargs)
        if isinstance(path, str) and os.path.exists(path):
            workspace = arpes.config.CONFIG['WORKSPACE']

            try:
                workspace = workspace['name']
            except (TypeError, KeyError):
                pass

            if not workspace or workspace not in path:
                warnings.warn(('Plotting function {} appears not to abide by '
                               'practice of placing plots into designated workspaces.').format(plot_fn.__name__))

            provenance_context = {
                'VERSION': VERSION,
                'time': datetime.datetime.now().isoformat(),
                'name': plot_fn.__name__,
                'args': [arg.attrs.get('provenance', {}) for arg in args
                         if isinstance(arg, xr.DataArray)],
                'kwargs': {k: v.attrs.get('provenance', {}) for k, v in kwargs.items()
                           if isinstance(v, xr.DataArray)}
            }

            provenance_path = path + '.provenance.json'
            with open(provenance_path, 'w') as f:
                json.dump(provenance_context, f, indent=2)

        return path

    return func_wrapper


def provenance(child_arr: xr.DataArray, parent_arr: typing.Union[DataType, typing.List[DataType]], record, keep_parent_ref=False):
    """
    Function that updates the provenance for a piece of data with a single parent.

    :param child_arr:
    :param parent_arr:
    :param record:
    :param keep_parent_ref:
    :return:
    """
    if isinstance(parent_arr, list):
        assert len(parent_arr) == 1
        parent_arr = parent_arr[0]

    if 'id' not in child_arr.attrs:
        attach_id(child_arr)

    parent_id = parent_arr.attrs.get('id')
    if parent_id is None:
        warnings.warn('Parent array has no ID.')

    if child_arr.attrs['id'] == parent_id:
        warnings.warn('Duplicate id for dataset %s' % child_arr.attrs['id'])

    child_arr.attrs['provenance'] = {
        'record': record,
        'parent_id': parent_id,
        'parents_provanence': parent_arr.attrs.get('provenance'),
        'time': datetime.datetime.now().isoformat(),
        'version': VERSION,
    }

    if keep_parent_ref:
        child_arr.attrs['provenance']['parent'] = parent_arr


def provenance_multiple_parents(child_arr: xr_types, parents, record, keep_parent_ref=False):
    """
    Similar to `provenance` updates the data provenance information for data with
    multiple sources or "parents". For instance, if you normalize a piece of data "X" by a metal
    reference "Y", then the returned data would list both "X" and "Y" in its history.

    :param child_arr:
    :param parents:
    :param record:
    :param keep_parent_ref:
    :return:
    """
    if 'id' not in child_arr.attrs:
        attach_id(child_arr)

    if child_arr.attrs['id'] in {p.attrs.get('id', None) for p in parents}:
        warnings.warn('Duplicate id for dataset %s' % child_arr.attrs['id'])

    child_arr.attrs['provenance'] = {
        'record': record,
        'parent_id': [p.attrs['id'] for p in parents],
        'parents_provenance': [p.attrs['provenance'] for p in parents],
        'time': datetime.datetime.now().isoformat(),
    }

    if keep_parent_ref:
        child_arr.attrs['provenance']['parent'] = parents



import numpy as np

import xarray as xr
from typing import Any

__all__ = ['CoordinateConverter', 'K_SPACE_BORDER', 'MOMENTUM_BREAKPOINTS']

K_SPACE_BORDER = 0.02
MOMENTUM_BREAKPOINTS = [0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1]


class CoordinateConverter:
    def __init__(self, arr: xr.DataArray, dim_order=None, *args, **kwargs):
        # Intern the volume so that we can check on things during computation
        self.arr = arr
        self.dim_order = dim_order

    def prep(self, arr: xr.DataArray):
        """
        The CoordinateConverter.prep method allows you to pre-compute some transformations
        that are common to the individual coordinate transform methods as an optimization.

        This is useful if you want the conversion methods to have separation of concern,
        but if it is advantageous for them to be able to share a computation of some
        variable. An example of this is in BE-kx-ky conversion, where computing k_p_tot
        is a step in both converting kx and ky, so we would like to avoid doing it twice.

        Of course, you can neglect this function entirely. Another technique is to simple
        cache computations as they arrive. This is the technique that is used in
        ConvertKxKy below
        :param arr:
        :return:
        """
        pass

    @property
    def is_slit_vertical(self) -> bool:
        # 89 - 91 degrees
        return np.abs(self.arr.S.lookup_offset_coord('alpha') - np.pi / 2) < (np.pi / 180)

    def kspace_to_BE(self, binding_energy: np.ndarray, *args: np.ndarray, **kwargs: Any) -> np.ndarray:
        return binding_energy

    def conversion_for(self, dim):
        pass

    def identity_transform(self, axis_name, *args, **kwargs):
        return args[self.dim_order.index(axis_name)]

    def get_coordinates(self, resolution: dict=None):
        coordinates = {}
        coordinates['eV'] = self.arr.coords['eV']
        return coordinates



import matplotlib.pyplot as plt

import xarray as xr
from arpes.provenance import save_plot_provenance

__all__ = ('plot_parameter',)


@save_plot_provenance
def plot_parameter(fit_data: xr.DataArray, param_name: str, ax=None,
                   fillstyle='none',
                   shift=0, x_shift=0,
                   markersize=8, title=None, out=None, two_sigma=False, **kwargs):
    fig = None

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get('figsize', (7, 5)))

    ds = fit_data.F.param_as_dataset(param_name)
    x = ds.coords[ds.value.dims[0]].values
    print(kwargs)
    color = kwargs.get('color')
    e_width = None
    l_width = None
    if two_sigma:
        line, markers, lines = ax.errorbar(x + x_shift, ds.value.values + shift, yerr=2 * ds.error.values, fmt='', elinewidth=1, linewidth=0, c=color)
        color = lines[0].get_color()[0]
        e_width = 2
        l_width = 0

    ax.errorbar(x + x_shift, ds.value.values + shift, yerr=ds.error.values, fmt='s',
                color=color, elinewidth=e_width, linewidth=l_width,
                markeredgewidth=e_width or 2,
                fillstyle=fillstyle,
                markersize=markersize)



import lmfit as lf
import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.fits import (GStepBModel, LinearModel, QuadraticModel,
                        broadcast_model)
from arpes.provenance import provenance, update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.math import shift_by


def _exclude_from_set(excluded):
    def exclude(l):
        return list(set(l).difference(excluded))

    return exclude


exclude_hemisphere_axes = _exclude_from_set({'phi', 'eV'})
exclude_hv_axes = _exclude_from_set({'hv', 'eV'})


__all__ = (
    'build_quadratic_fermi_edge_correction',
    'build_photon_energy_fermi_edge_correction', 'apply_photon_energy_fermi_edge_correction',
    'apply_quadratic_fermi_edge_correction', 'apply_copper_fermi_edge_correction',
    'apply_direct_copper_fermi_edge_correction', 'build_direct_fermi_edge_correction',
    'apply_direct_fermi_edge_correction', 'find_e_fermi_linear_dos',
)


def find_e_fermi_linear_dos(edc, guess=None, plot=False, ax=None):
    """
    Does a reasonable job of finding E_Fermi in-situ for graphene/graphite or other materials with a linear DOS near
    the chemical potential. You can provide an initial guess via guess, or one will be chosen half way through the EDC.

    The Fermi level is estimated as the location where the DoS crosses below an estimated background level
    :param edc:
    :param guess:
    :param plot:
    :return:
    """
    if guess is None:
        guess = edc.eV.values[len(edc.eV) // 2]

    edc = edc - np.percentile(edc.values, (20,))[0]
    mask = edc > np.percentile(edc.sel(eV=slice(None, guess)), 20)
    mod = LinearModel().guess_fit(edc[mask])

    chemical_potential = -mod.params['intercept'].value / mod.params['slope'].value

    if plot:
        if ax is None:
            fig, ax = plt.subplots()
        edc.plot(ax=ax)
        ax.axvline(chemical_potential, linestyle='--', color='red')
        ax.axvline(guess, linestyle='--', color='gray')

    return chemical_potential


@update_provenance('Apply direct Fermi edge correction from metal reference')
def apply_direct_copper_fermi_edge_correction(arr: DataType, copper_ref: DataType, *args, **kwargs):
    """
    Applies a *direct* fermi edge correction.
    :param arr:
    :param copper_ref:
    :param args:
    :param kwargs:
    :return:
    """
    arr = normalize_to_spectrum(arr)
    copper_ref = normalize_to_spectrum(copper_ref)
    direct_corr = build_direct_fermi_edge_correction(copper_ref, *args, **kwargs)
    shift = np.interp(arr.coords['phi'].values, direct_corr.coords['phi'].values,
                      direct_corr.values)
    return apply_direct_fermi_edge_correction(arr, shift)


def apply_direct_fermi_edge_correction(arr: xr.DataArray, correction=None, *args, **kwargs):
    if correction is None:
        correction = build_direct_fermi_edge_correction(arr, *args, **kwargs)

    shift_amount = -correction / arr.T.stride(generic_dim_names=False)['eV']  # pylint: disable=invalid-unary-operand-type
    energy_axis = list(arr.dims).index('eV')

    correction_axis = list(arr.dims).index(correction.dims[0])

    corrected_arr = xr.DataArray(
        shift_by(arr.values, shift_amount, axis=energy_axis, by_axis=correction_axis, order=1),
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if 'id'in corrected_arr.attrs:
        del corrected_arr.attrs['id']

    provenance(corrected_arr, arr, {
        'what': 'Shifted Fermi edge to align at 0 along hv axis',
        'by': 'apply_photon_energy_fermi_edge_correction',
        'correction': list(correction.values if isinstance(correction, xr.DataArray) else correction),
    })

    return corrected_arr


@update_provenance('Build direct Fermi edge correction')
def build_direct_fermi_edge_correction(arr: xr.DataArray, fit_limit=0.001, energy_range=None, plot=False,
                                       along='phi'):
    """
    Builds a direct fermi edge correction stencil.

    This means that fits are performed at each value of the 'phi' coordinate
    to get a list of fits. Bad fits are thrown out to form a stencil.

    This can be used to shift coordinates by the nearest value in the stencil.

    :param copper_ref:
    :param args:
    :param kwargs:
    :return:
    """

    if energy_range is None:
        energy_range = slice(-0.1, 0.1)

    exclude_axes = ['eV', along]
    others = [d for d in arr.dims if d not in exclude_axes]
    edge_fit = broadcast_model(GStepBModel, arr.sum(others).sel(eV=energy_range), along).results

    def sieve(c, v):
        return v.item().params['center'].stderr < 0.001

    corrections = edge_fit.T.filter_coord(along, sieve).T.map(lambda x: x.params['center'].value)

    if plot:
        corrections.plot()

    return corrections


@update_provenance('Apply quadratic Fermi edge correction from metal reference')
def apply_copper_fermi_edge_correction(arr: DataType, copper_ref: DataType, *args, **kwargs):
    # this maybe isn't best because we don't correct anything other than the spectrum,
    # but that's the only thing with an energy axis in ARPES datasets so whatever
    arr = normalize_to_spectrum(arr)
    copper_ref = normalize_to_spectrum(copper_ref)
    quadratic_corr = build_quadratic_fermi_edge_correction(copper_ref, *args, **kwargs)
    return apply_quadratic_fermi_edge_correction(arr, quadratic_corr)


def build_quadratic_fermi_edge_correction(arr: xr.DataArray, fit_limit=0.001, eV_slice=None, plot=False) -> lf.model.ModelResult:
    # TODO improve robustness here by allowing passing in the location of the fermi edge guess
    # We could also do this automatically by using the same method we use for step detection to find the edge of the
    # spectrometer image

    if eV_slice is None:
        approximate_fermi_level = arr.S.find_spectrum_energy_edges().max()
        eV_slice = slice(approximate_fermi_level-0.4, approximate_fermi_level+0.4)
    else:
        approximate_fermi_level = 0
    sum_axes = exclude_hemisphere_axes(arr.dims)
    edge_fit = broadcast_model(GStepBModel, arr.sum(sum_axes).sel(eV=eV_slice), 'phi', params={'center': {'value': approximate_fermi_level}})

    size_phi = len(arr.coords['phi'])
    not_nanny = (np.logical_not(np.isnan(arr)) * 1).sum('eV') > size_phi * 0.30
    condition = np.logical_and(edge_fit.F.s('center') < fit_limit, not_nanny)

    quadratic_corr = QuadraticModel().guess_fit(
        edge_fit.F.p('center'),
        weights=condition * 1)
    if plot:
        edge_fit.F.p('center').plot()
        plt.plot(arr.coords['phi'], quadratic_corr.best_fit)

    return quadratic_corr


@update_provenance('Build photon energy Fermi edge correction')
def build_photon_energy_fermi_edge_correction(arr: xr.DataArray, plot=False, energy_window=0.2):
    edge_fit = broadcast_model(GStepBModel, arr.sum(exclude_hv_axes(arr.dims)).sel(
        eV=slice(-energy_window, energy_window)), 'hv')

    return edge_fit


def apply_photon_energy_fermi_edge_correction(arr: xr.DataArray, correction=None, **kwargs):
    if correction is None:
        correction = build_photon_energy_fermi_edge_correction(arr, **kwargs)

    correction_values = correction.T.map(lambda x: x.params['center'].value)
    if 'corrections' not in arr.attrs:
        arr.attrs['corrections'] = {}

    arr.attrs['corrections']['hv_correction'] = list(correction_values.values)

    shift_amount = -correction_values / arr.T.stride(generic_dim_names=False)['eV']
    energy_axis = arr.dims.index('eV')
    hv_axis = arr.dims.index('hv')

    corrected_arr = xr.DataArray(
        shift_by(arr.values, shift_amount, axis=energy_axis, by_axis=hv_axis, order=1),
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if 'id' in corrected_arr.attrs:
        del corrected_arr.attrs['id']

    provenance(corrected_arr, arr, {
        'what': 'Shifted Fermi edge to align at 0 along hv axis',
        'by': 'apply_photon_energy_fermi_edge_correction',
        'correction': list(correction_values.values),
    })

    return corrected_arr


def apply_quadratic_fermi_edge_correction(arr: xr.DataArray, correction: lf.model.ModelResult=None, offset=None):
    assert isinstance(arr, xr.DataArray)
    if correction is None:
        correction = build_quadratic_fermi_edge_correction(arr)

    if 'corrections' not in arr.attrs:
        arr.attrs['corrections'] = {}

    arr.attrs['corrections']['FE_Corr'] = correction.best_values

    delta_E = arr.coords['eV'].values[1] - arr.coords['eV'].values[0]
    dims = list(arr.dims)
    energy_axis = dims.index('eV')
    phi_axis = dims.index('phi')

    shift_amount_E = correction.eval(x=arr.coords['phi'].values)

    if offset is not None:
        shift_amount_E = shift_amount_E - offset

    shift_amount = -shift_amount_E / delta_E

    corrected_arr = xr.DataArray(
        shift_by(arr.values, shift_amount, axis=energy_axis, by_axis=phi_axis, order=1),
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )

    if 'id' in corrected_arr.attrs:
        del corrected_arr.attrs['id']

    provenance(corrected_arr, arr, {
        'what': 'Shifted Fermi edge to align at 0',
        'by': 'apply_quadratic_fermi_edge_correction',
        'correction': correction.best_values,
    })

    return corrected_arr



import xarray as xr
from arpes.provenance import update_provenance
from arpes.utilities import lift_dataarray_to_generic

__all__ = ('mean_and_deviation',)


@update_provenance('Calculate mean and standard deviation for observation axis')
@lift_dataarray_to_generic
def mean_and_deviation(data: xr.DataArray, axis=None, name=None):
    """
    Calculates the mean and standard deviation of a DataArray along an axis
    that corresponds to individual observations. This axis can be passed or
    inferred from a set of standard observation-like axes.

    New data variables are created with names `{name}` and `{name}_std`.
    If a name is not attached to the DataArray, it should be provided.

    :param data:
    :param axis:
    :param name:
    :return:
    """
    preferred_axes = ['bootstrap', 'cycle', 'idx']

    name = data.name if data.name is not None else name
    assert name is not None

    if axis is None:
        for pref_axis in preferred_axes:
            if pref_axis in data.dims:
                axis = pref_axis
                break

    assert axis in data.dims
    return xr.Dataset(data_vars={name: data.mean(axis), name + '_std': data.std(axis)}, attrs=data.attrs)



"""
Provides a jumping off point for defining data loading plugins using the NeXuS file format.
Currently we assume that the raw file format is actually HDF.
"""

__all__ = ('read_data_attributes_from',)


def read_data_attributes_from(group, paths):
    """
    Reads simple (float, string, int, etc) leaves from
    a nested description of paths out of a NeXuS file.
    :param group:
    :param paths:
    :return:
    """
    read_attrs = {}
    original_group = group
    for path, attributes in paths:
        group = original_group
        for p in path:
            group = group[p]

        for attribute_name in attributes:
            try:
                data = group[attribute_name]['data']
            except ValueError:
                data = group[attribute_name]
            try:
                data = data[:]
            except ValueError:
                data = data.value

            read_attrs[attribute_name] = data

    return read_attrs



import numpy as np

import arpes.constants
import xarray as xr
from typing import Any, Callable, List

from .base import CoordinateConverter, K_SPACE_BORDER, MOMENTUM_BREAKPOINTS
from .bounds_calculations import calculate_kp_bounds, calculate_kx_ky_bounds

__all__ = ['ConvertKp', 'ConvertKxKy']

class ConvertKp(CoordinateConverter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.k_tot = None

    def get_coordinates(self, resolution: dict=None):
        if resolution is None:
            resolution = {}

        coordinates = super().get_coordinates(resolution)
        (kp_low, kp_high) = calculate_kp_bounds(self.arr)

        inferred_kp_res = (kp_high - kp_low + 2 * K_SPACE_BORDER) / len(self.arr.coords['phi'])
        inferred_kp_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kp_res][-1]

        coordinates['kp'] = np.arange(kp_low - K_SPACE_BORDER, kp_high + K_SPACE_BORDER,
                                      resolution.get('kp', inferred_kp_res))

        base_coords = {k: v for k, v in self.arr.coords.items()
                       if k not in ['eV', 'phi', 'beta', 'theta']}

        coordinates.update(base_coords)
        return coordinates

    def compute_k_tot(self, binding_energy: np.ndarray) -> None:
        self.k_tot = arpes.constants.K_INV_ANGSTROM * np.sqrt(
            self.arr.S.hv - self.arr.S.work_function + binding_energy)

    def kspace_to_phi(self, binding_energy: np.ndarray, kp: np.ndarray, *args: Any, **kwargs: Any) -> np.ndarray:
        # check signs again here
        if self.is_slit_vertical:
            polar_angle = self.arr.S.lookup_offset_coord('theta') + self.arr.S.lookup_offset_coord('psi')
            parallel_angle = self.arr.S.lookup_offset_coord('beta')
        else:
            polar_angle = self.arr.S.lookup_offset_coord('beta') + self.arr.S.lookup_offset_coord('psi')
            parallel_angle = self.arr.S.lookup_offset_coord('theta')

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        # TODO verify this
        return np.arcsin(kp / self.k_tot / np.cos(polar_angle)) + self.arr.S.phi_offset + parallel_angle

    def conversion_for(self, dim: str) -> Callable:
        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            'eV': self.kspace_to_BE,
            'phi': self.kspace_to_phi
        }.get(dim, with_identity)


class ConvertKxKy(CoordinateConverter):
    """
    Please note that currently we assume that psi = 0 when you are not using an
    electrostatic deflector
    """
    def __init__(self, arr: xr.DataArray, *args: List[str], **kwargs: Any) -> None:
        super().__init__(arr, *args, **kwargs)
        self.k_tot = None
        self.phi = None
        # the angle perpendicular to phi as appropriate to the scan, this can be any of
        # psi, theta, beta
        self.perp_angle = None

        self.rkx = None
        self.rky = None

        # accept either vertical or horizontal, fail otherwise
        if not any(np.abs(arr.alpha - alpha_option) < (np.pi / 180) for alpha_option in [0, np.pi/2]):
            raise ValueError('You must convert either vertical or horizontal slit data with this converter.')

        self.direct_angles = ('phi', [d for d in ['psi', 'beta', 'theta'] if d in arr.indexes][0])

        if self.direct_angles[1] != 'psi':
            # psi allows for either orientation
            assert (self.direct_angles[1] in {'theta'}) != (not self.is_slit_vertical)

        # determine which other angles constitute equivalent sets
        if self.is_slit_vertical:
            self.parallel_angles = ('beta', 'theta')
        else:
            self.parallel_angles = ('theta', 'beta')

    def get_coordinates(self, resolution: dict = None):
        if resolution is None:
            resolution = {}

        coordinates = super().get_coordinates(resolution)

        ((kx_low, kx_high), (ky_low, ky_high)) = calculate_kx_ky_bounds(self.arr)

        kx_angle, ky_angle = self.direct_angles
        if self.is_slit_vertical:
            # phi actually measures along ky
            ky_angle, kx_angle = kx_angle, ky_angle

        len_ky_angle = len(self.arr.coords[ky_angle])
        len_kx_angle = len(self.arr.coords[kx_angle])

        inferred_kx_res = (kx_high - kx_low + 2 * K_SPACE_BORDER) / len(self.arr.coords[kx_angle])
        inferred_ky_res = (ky_high - ky_low + 2 * K_SPACE_BORDER) / len(self.arr.coords[ky_angle])

        # upsample a bit if there aren't that many points along a certain axis
        try:
            inferred_kx_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kx_res][-2 if (len_kx_angle < 80) else -1]
        except IndexError:
            inferred_kx_res = MOMENTUM_BREAKPOINTS[-2]
        try:
            inferred_ky_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_ky_res][-2 if (len_ky_angle < 80) else -1]
        except IndexError:
            inferred_ky_res = MOMENTUM_BREAKPOINTS[-2]

        coordinates['kx'] = np.arange(kx_low - K_SPACE_BORDER, kx_high + K_SPACE_BORDER,
                                      resolution.get('kx', inferred_kx_res))
        coordinates['ky'] = np.arange(ky_low - K_SPACE_BORDER, ky_high + K_SPACE_BORDER,
                                      resolution.get('ky', inferred_ky_res))

        base_coords = {k: v for k, v in self.arr.coords.items()
                       if k not in ['eV', 'phi', 'psi', 'theta', 'beta', 'alpha', 'chi']}
        coordinates.update(base_coords)

        return coordinates

    def compute_k_tot(self, binding_energy: np.ndarray) -> None:
        self.k_tot = arpes.constants.K_INV_ANGSTROM * np.sqrt(
            self.arr.S.hv - self.arr.S.work_function + binding_energy)

    def conversion_for(self, dim: str) -> Callable:
        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            'eV': self.kspace_to_BE,
            'phi': self.kspace_to_phi,
            'theta': self.kspace_to_perp_angle,
            'psi': self.kspace_to_perp_angle,
            'beta': self.kspace_to_perp_angle,
        }.get(dim, with_identity)

    @property
    def needs_rotation(self) -> bool:
        # force rotation when greater than 0.5 deg
        return np.abs(self.arr.S.lookup_offset_coord('chi')) > (0.5 * np.pi / 180)

    def rkx_rky(self, kx, ky):
        """
        Returns the rotated kx and ky values when we are rotating by nonzero chi
        :return:
        """

        if self.rkx is not None:
            return self.rkx, self.rky

        chi = self.arr.S.lookup_offset_coord('chi')

        self.rkx = kx * np.cos(chi) - ky * np.sin(chi)
        self.rky = ky * np.cos(chi) + kx * np.sin(chi)

        return self.rkx, self.rky

    def kspace_to_phi(self, binding_energy: np.ndarray, kx: np.ndarray, ky: np.ndarray, *args: Any, **kwargs: Any) -> np.ndarray:
        if self.phi is not None:
            return self.phi

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        if self.needs_rotation:
            kx, ky = self.rkx_rky(kx, ky)

        # This can be condensed but it is actually better not to condense it:
        # In this format, we can very easily compare to the raw coordinate conversion functions that
        # come from Mathematica in order to adjust signs, etc.
        scan_angle = self.direct_angles[1]
        if scan_angle == 'psi':
            if self.is_slit_vertical:
                self.phi = np.arcsin(ky / np.sqrt(self.k_tot ** 2 - kx ** 2)) + self.arr.S.phi_offset + \
                           self.arr.S.lookup_offset_coord(self.parallel_angles[0])
            else:
                self.phi = np.arcsin(kx / np.sqrt(self.k_tot ** 2 - ky ** 2)) + self.arr.S.phi_offset + \
                           self.arr.S.lookup_offset_coord(self.parallel_angles[0])
        elif scan_angle == 'beta':
            # vertical slit
            self.phi = np.arcsin(kx / self.k_tot) + self.arr.S.phi_offset + \
                       self.arr.S.lookup_offset_coord(self.parallel_angles[0])
        elif scan_angle == 'theta':
            # vertical slit
            self.phi = np.arcsin(ky / self.k_tot) + self.arr.S.phi_offset + \
                       self.arr.S.lookup_offset_coord(self.parallel_angles[0])
        else:
            raise ValueError('No recognized scan angle found for {}'.format(self.parallel_angles[1]))

        return self.phi

    def kspace_to_perp_angle(self, binding_energy: np.ndarray, kx: np.ndarray, ky: np.ndarray, *args: Any, **kwargs: Any) -> np.ndarray:
        if self.perp_angle is not None:
            return self.perp_angle

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        if self.needs_rotation:
            kx, ky = self.rkx_rky(kx, ky)

        scan_angle = self.direct_angles[1]
        if scan_angle == 'psi':
            if self.is_slit_vertical:
                self.perp_angle = -np.arcsin(kx / self.k_tot) + self.arr.S.psi_offset - \
                                  self.arr.S.lookup_offset_coord(self.parallel_angles[1])
            else:
                self.perp_angle = np.arcsin(ky / self.k_tot) + self.arr.S.psi_offset + \
                                  self.arr.S.lookup_offset_coord(self.parallel_angles[1])
        elif scan_angle == 'beta':
            self.perp_angle = -np.arcsin(ky / np.sqrt(self.k_tot ** 2 - kx ** 2)) + self.arr.S.beta_offset
            pass
        elif scan_angle == 'theta':
            self.perp_angle = -np.arcsin(kx / np.sqrt(self.k_tot ** 2 - ky ** 2)) + self.arr.S.theta_offset
            pass
        else:
            raise ValueError('No recognized scan angle found for {}'.format(self.parallel_angles[1]))

        return self.perp_angle



import itertools
from collections import defaultdict

import numpy as np

import arpes.constants
import arpes.models.band
import arpes.utilities
import arpes.utilities.math
import typing
import xarray as xr
from arpes.fits import GStepBModel, broadcast_model
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.math import fermi_distribution

from .filters import gaussian_filter_arr

__all__ = ('normalize_by_fermi_distribution', 'symmetrize_axis', 'condense', 'rebin',
           'fit_fermi_edge')


@update_provenance('Fit Fermi Edge')
def fit_fermi_edge(data, energy_range=None):
    """
    Fits a Fermi edge. Not much easier than doing it manually, but this can be
    useful sometimes inside procedures where you don't want to reimplement this logic.
    :param data:
    :param energy_range:
    :return:
    """
    if energy_range is None:
        energy_range = slice(-0.1, 0.1)

    broadcast_directions = list(data.dims)
    broadcast_directions.remove('eV')
    assert len(broadcast_directions) == 1 # for now we don't support more

    edge_fit = broadcast_model(GStepBModel, data.sel(eV=energy_range), broadcast_directions[0])
    return edge_fit


@update_provenance('Normalized by the 1/Fermi Dirac Distribution at sample temp')
def normalize_by_fermi_distribution(
        data: DataType, max_gain=None, rigid_shift=0, instrumental_broadening=None, total_broadening=None):
    """
    Normalizes a scan by 1/the fermi dirac distribution. You can control the maximum gain with ``clamp``, and whether
    the Fermi edge needs to be shifted (this is for those desperate situations where you want something that
    "just works") via ``rigid_shift``.

    :param data: Input
    :param max_gain: Maximum value for the gain. By default the value used is the mean of the spectrum.
    :param rigid_shift: How much to shift the spectrum chemical potential.
    Pass the nominal value for the chemical potential in the scan. I.e. if the chemical potential is at BE=0.1, pass
    rigid_shift=0.1.
    :param instrumental_broadening: Instrumental broadening to use for convolving the distribution
    :return: Normalized DataArray
    """
    data = normalize_to_spectrum(data)

    if total_broadening:
        distrib = fermi_distribution(data.coords['eV'].values - rigid_shift,
                                     total_broadening / arpes.constants.K_BOLTZMANN_EV_KELVIN)
    else:
        distrib = fermi_distribution(data.coords['eV'].values - rigid_shift, data.S.temp)

    # don't boost by more than 90th percentile of input, by default
    if max_gain is None:
        max_gain = min(np.mean(data.values), np.percentile(data.values, 10))

    distrib[distrib < 1/max_gain] = 1/max_gain
    distrib_arr = xr.DataArray(
        distrib,
        {'eV': data.coords['eV'].values},
        ['eV']
    )

    if instrumental_broadening is not None:
        distrib_arr = gaussian_filter_arr(distrib_arr, sigma={'eV': instrumental_broadening})

    return data / distrib_arr


@update_provenance('Symmetrize about axis')
def symmetrize_axis(data, axis_name, flip_axes=None, shift_axis=True):
    """
    Symmetrizes data across an axis. It would be better ultimately to be able
    to implement an arbitrary symmetry (such as a mirror or rotational symmetry
    about a line or point) and to symmetrize data by that method.

    :param data:
    :param axis_name:
    :param flip_axes:
    :param shift_axis:
    :return:
    """
    data = data.copy(deep=True) # slow but make sure we don't bork axis on original
    data.coords[axis_name].values = data.coords[axis_name].values - data.coords[axis_name].values[0]

    selector = {}
    selector[axis_name] = slice(None, None, -1)
    rev = data.sel(**selector).copy()

    rev.coords[axis_name].values = -rev.coords[axis_name].values

    if flip_axes is None:
        flip_axes = []

    for axis in flip_axes:
        selector = {}
        selector[axis] = slice(None, None, -1)
        rev = rev.sel(**selector)
        rev.coords[axis].values = -rev.coords[axis].values

    return rev.combine_first(data)


@update_provenance('Condensed array')
def condense(data: xr.DataArray):
    """
    Clips the data so that only regions where there is substantial weight are included. In
    practice this usually means selecting along the ``eV`` axis, although other selections
    might be made.

    :param data: xarray.DataArray
    :return:
    """
    if 'eV' in data.dims:
        data = data.sel(eV=slice(None, 0.05))

    return data


@update_provenance('Rebinned array')
def rebin(data: DataType, shape: dict=None, reduction: typing.Union[int, dict]=None, interpolate=False, **kwargs):
    """
    Rebins the data onto a different (smaller) shape. By default the behavior is to
    split the data into chunks that are integrated over. An interpolation option is also
    available.

    Exactly one of ``shape`` and ``reduction`` should be supplied.

    Dimensions corresponding to missing entries in ``shape`` or ``reduction`` will not
    be changed.

    :param data:
    :param interpolate: Use interpolation instead of integration
    :param shape: Target shape
    :param reduction: Factor to reduce each dimension by
    :return:
    """

    if isinstance(data, xr.Dataset):
        new_vars = {
            datavar: rebin(data[datavar], shape=shape, reduction=reduction, interpolate=interpolate, **kwargs)
            for datavar in data.data_vars
        }
        new_coords = {}

        for var in new_vars.values():
            new_coords.update(var.coords)
        return xr.Dataset(data_vars=new_vars, coords=new_coords, attrs=data.attrs)

    data = arpes.utilities.normalize_to_spectrum(data)

    if any(d in kwargs for d in data.dims):
        reduction = kwargs

    if interpolate:
        raise NotImplementedError('The interpolation option has not been implemented')

    assert(shape is None or reduction is None)

    if isinstance(reduction, int):
        reduction = {d: reduction for d in data.dims}

    if reduction is None:
        reduction = {}

    # we standardize by computing reduction from shape is shape was supplied.
    if shape is not None:
        reduction = {k: len(data.coords[k]) // v for k, v in shape.items()}

    # since we are not interpolating, we need to clip each dimension so that the reduction
    # factor evenly divides the real shape of the input data.
    slices = defaultdict(lambda: slice(None))

    if not data.dims:
        return data

    for dim, reduction_factor in reduction.items():
        remainder = len(data.coords[dim]) % reduction_factor
        if remainder != 0:
            slices[dim] = slice(None, -remainder)

    trimmed_data = data.data[[slices[d] for d in data.dims]]
    trimmed_coords = {d: coord[slices[d]] for d, coord in data.indexes.items()}

    temp_shape = [[trimmed_data.shape[i] // reduction.get(d, 1), reduction.get(d, 1)]
                  for i, d in enumerate(data.dims)]
    temp_shape = itertools.chain(*temp_shape)
    reduced_data = trimmed_data.reshape(*temp_shape)

    for i in range(len(data.dims)):
        reduced_data = reduced_data.mean(i + 1)

    reduced_coords = {d: coord[::reduction.get(d, 1)] for d, coord in trimmed_coords.items()}

    return xr.DataArray(
        reduced_data,
        reduced_coords,
        data.dims,
        attrs=data.attrs
    )



from enum import Enum

from typing import Union

__all__ = ['REGIONS', 'normalize_region', 'DesignatedRegions']


class DesignatedRegions(Enum):
    # angular windows
    NARROW_ANGLE = 0 # trim to a narrow central region in the spectrometer
    WIDE_ANGLE = 1 # trim to just inside edges of spectrometer data
    TRIM_EMPTY = 2 # trim to edges of spectrometer data

    # energy windows
    BELOW_EF = 10  # everything below e_F
    ABOVE_EF = 11  # everything above e_F
    EF_NARROW = 12  # narrow cut around e_F
    MESO_EF = 13 # comfortably below e_F, pun on mesosphere

    # effective energy windows, determined by Canny edge detection
    BELOW_EFFECTIVE_EF = 20  # everything below e_F
    ABOVE_EFFECTIVE_EF = 21  # everything above e_F
    EFFECTIVE_EF_NARROW = 22  # narrow cut around e_F
    MESO_EFFECTIVE_EF = 23 # comfortably below effective e_F, pun on mesosphere


REGIONS = {
    'copper_prior': {
        'eV': DesignatedRegions.MESO_EFFECTIVE_EF,
    },
    # angular can refer to either 'pixels' or 'phi'
    'wide_angular': {
        # angular can refer to either 'pixels' or 'phi'
        'angular': DesignatedRegions.WIDE_ANGLE,
    },
    'narrow_angular': {
        'angular': DesignatedRegions.NARROW_ANGLE,
    },
}


def normalize_region(region: Union[str, dict]):
    if isinstance(region, str):
        return REGIONS[region]

    if isinstance(region, dict):
        return region

    raise TypeError('Region should be either a string (i.e. an ID/alias) '
                    'or an explicit dictionary.')



import xarray as xr
from arpes.typing import DataType

__all__ = ('apply_dataarray', 'lift_datavar_attrs', 'lift_dataarray_attrs', 'lift_dataarray',)

def apply_dataarray(arr: xr.DataArray, f, *args, **kwargs):
    return xr.DataArray(
        f(arr.values, *args, **kwargs),
        arr.coords,
        arr.dims,
        attrs=arr.attrs
    )


def lift_dataarray(f):
    """
    Lifts a function that operates on an np.ndarray's values to one that
    acts on the values of an xr.DataArray
    :param f:
    :return: g: Function operating on an xr.DataArray
    """

    def g(arr: xr.DataArray, *args, **kwargs):
        return apply_dataarray(arr, f, *args, **kwargs)

    return g


def lift_dataarray_attrs(f):
    """
    Lifts a function that operates on a dictionary to a function that acts on the
    attributes of an xr.DataArray, producing a new xr.DataArray. Another option
    if you don't need to create a new DataArray is to modify the attributes.
    :param f:
    :return: g: Function operating on the attributes of an xr.DataArray
    """

    def g(arr: xr.DataArray, *args, **kwargs):
        return xr.DataArray(
            arr.values,
            arr.coords,
            arr.dims,
            attrs=f(arr.attrs, *args, **kwargs)
        )

    return g


def lift_datavar_attrs(f):
    """
    Lifts a function that operates on a dictionary to a function that acts on the
    attributes of all the datavars in a xr.Dataset, as well as the Dataset attrs
    themselves.
    :param f: Function to apply
    :return:
    """

    def g(data: DataType, *args, **kwargs):
        arr_lifted = lift_dataarray_attrs(f)
        if isinstance(data, xr.DataArray):
            return arr_lifted(data, *args, **kwargs)

        var_names = list(data.data_vars.keys())
        new_vars = {k: arr_lifted(data[k], *args, **kwargs) for k in var_names}
        new_root_attrs = f(data.attrs, *args, **kwargs)

        return xr.Dataset(new_vars, data.coords, new_root_attrs)

    return g



import copy
import functools
import itertools

import numpy as np
from scipy.spatial import distance

import arpes.models.band
import arpes.utilities.math
import xarray as xr
from arpes.constants import HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ
from arpes.fits import (AffineBackgroundModel, LorentzianModel, QuadraticModel,
                        broadcast_model)
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import enumerate_dataarray, normalize_to_spectrum
from arpes.utilities.conversion.forward import \
    convert_coordinates_to_kspace_forward
from arpes.utilities.jupyter_utils import wrap_tqdm

__all__ = ('fit_bands', 'fit_for_effective_mass',)


def fit_for_effective_mass(data: DataType, fit_kwargs=None):
    """
    Performs an effective mass fit by first fitting for Lorentzian lineshapes and then fitting a quadratic
    model to the result. This is an alternative to global effective mass fitting.

    In the case that data is provided in anglespace, the Lorentzian fits are performed in anglespace
    before being converted to momentum where the effective mass is extracted.

    We should probably include uncertainties here.

    :param data:
    :param fit_kwargs: Passthrough for arguments to `broadcast_model`,
    used internally to obtain the Lorentzian peak locations
    :return:
    """
    if fit_kwargs is None:
        fit_kwargs = {}
    data = normalize_to_spectrum(data)
    mom_dim = [d for d in ['kp', 'kx', 'ky', 'kz', 'phi', 'beta', 'theta'] if d in data.dims][0]

    results = broadcast_model([LorentzianModel, AffineBackgroundModel], data, mom_dim, **fit_kwargs)
    if mom_dim in {'phi', 'beta', 'theta'}:
        forward = convert_coordinates_to_kspace_forward(data)
        final_mom = [d for d in ['kx', 'ky', 'kp', 'kz'] if d in forward][0]
        eVs = results.F.p('a_center').values
        kps = [forward[final_mom].sel(eV=eV, **dict([[mom_dim, ang]]), method='nearest') for
               eV, ang in zip(eVs, data.coords[mom_dim].values)]
        quad_fit = QuadraticModel().fit(eVs, x=np.array(kps))

        return HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ / (2 * quad_fit.params['a'].value)

    quad_fit = QuadraticModel().guess_fit(results.F.p('a_center'))
    return HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ / (2 * quad_fit.params['a'].value)


def unpack_bands_from_fit(band_results: xr.DataArray, weights=None, use_stderr_weighting=True):
    """
    This function is used to deconvolve the band identities of a series of overlapping bands.
    Sometimes through the fitting process, or across a place in the band structure where there is a nodal
    point, the identities of the bands across sequential fits can get mixed up.

    We can try to restore this identity by using the cosine similarity of fits, where the fit is represented
    as a vector by:

        v_band =  (sigma, amplitude, center) * weights
        weights = (5, 1/5, 10)

    For any point in the band structure, we find the closest place where we have fixed the band identities.
    Let the bands be indexed by i so that the bands are b_i and b_i_0 at the point of interest and at the reference
    respectively.

    Then, we calculate the matrix:
        s_ij = sim(b_i, b_j_0)

    The band identities are subsequently chosen so that the trace of this matrix is maximized among possible ways of
    labelling the bands b_i.

    The value of the weights parameter is chosen only to scale the dimensions so that they are closer to the
    same magnitude.

    :param arr:
    :param band_results:
    :param weights:
    :param use_stderr_weighting: Flag to indicate whether to scale vectors by the uncertainty
    :return:
    """
    if weights is None:
        weights = (2, 0, 10,)

    template_components = band_results.values[0].model.components
    prefixes = [component.prefix for component in template_components]

    identified_band_results = copy.deepcopy(band_results)

    def as_vector(model_fit, prefix=''):
        stderr = np.array([model_fit.params[prefix + 'sigma'].stderr,
                           model_fit.params[prefix + 'amplitude'].stderr,
                           model_fit.params[prefix + 'center'].stderr])
        return np.array([model_fit.params[prefix + 'sigma'].value,
                         model_fit.params[prefix + 'amplitude'].value,
                         model_fit.params[prefix + 'center'].value]) * weights / (1 + stderr)

    identified_by_coordinate = {}
    first_coordinate = None
    for coordinate, fit_result in enumerate_dataarray(band_results):
        frozen_coord = tuple(coordinate[d] for d in band_results.dims)

        closest_identified = None
        dist = float('inf')
        for coord, identified_band in identified_by_coordinate.items():
            current_dist = np.dot(coord, frozen_coord)
            if current_dist < dist:
                closest_identified = identified_band
                dist = current_dist

        if closest_identified is None:
            first_coordinate = coordinate
            closest_identified = [c.prefix for c in fit_result.model.components], fit_result
            identified_by_coordinate[frozen_coord] = closest_identified

        closest_prefixes, closest_fit = closest_identified
        mat_shape = (len(prefixes), len(prefixes),)
        dist_mat = np.zeros(shape=mat_shape)

        for i, j in np.ndindex(mat_shape):
            dist_mat[i, j] = distance.euclidean(
                as_vector(fit_result, prefixes[i]),
                as_vector(closest_fit, closest_prefixes[j])
            )

        best_arrangement = None
        best_trace = float('inf')
        for p in itertools.permutations(range(len(prefixes))):
            trace = sum(dist_mat[i, p_i] for i, p_i in enumerate(p))
            if trace < best_trace:
                best_trace = trace
                best_arrangement = p

        ordered_prefixes = [closest_prefixes[p_i] for p_i in best_arrangement]
        identified_by_coordinate[frozen_coord] = ordered_prefixes, fit_result
        identified_band_results.loc[coordinate] = ordered_prefixes

    # Now that we have identified the bands,
    # extract them into real bands
    bands = []
    for i in range(len(prefixes)):
        label = identified_band_results.loc[first_coordinate].values.item()[i]
        def dataarray_for_value(param_name, is_value):
            values = np.ndarray(shape=identified_band_results.values.shape, dtype=np.float)
            it = np.nditer(values, flags=['multi_index'], op_flags=['writeonly'])
            while not it.finished:
                prefix = identified_band_results.values[it.multi_index][i]
                param = band_results.values[it.multi_index].params[prefix + param_name]
                if is_value:
                    it[0] = param.value
                else:
                    it[0] = param.stderr
                it.iternext()

            return xr.DataArray(
                values,
                identified_band_results.coords,
                identified_band_results.dims
            )
        band_data = xr.Dataset({
            'center': dataarray_for_value('center', True),
            'center_stderr': dataarray_for_value('center', False),
            'amplitude': dataarray_for_value('amplitude', True),
            'amplitude_stderr': dataarray_for_value('amplitude', False),
            'sigma': dataarray_for_value('sigma', True),
            'sigma_stderr': dataarray_for_value('sigma', False),
        })
        bands.append(arpes.models.band.Band(label, data=band_data))

    return bands


@update_provenance('Fit bands from pattern')
def fit_patterned_bands(arr: xr.DataArray, band_set, direction_normal=True,
                        fit_direction=None, avoid_crossings=None,
                        stray=None, background=True, preferred_k_direction=None,
                        interactive=True, dataset=True):
    """
    Fits bands and determines dispersion in some region of a spectrum.

    The dimensions of the dataset are partitioned into three types:

    1. Fit directions, these are coordinates along the 1D (or maybe later 2D) marginals
    2. Broadcast directions, these are directions used to interpolate against the patterned directions
    3. Free directions, these are broadcasted but they are not used to extract initial values of the fit parameters

    For instance, if you laid out band patterns in a E, k_p, delay spectrum at delta_t=0, then if you are using MDCs,
    k_p is the fit direction, E is the broadcast direction, and delay is a free direction.

    In general we can recover the free directions and the broadcast directions implicitly by examining the band_set
    passed as a pattern.

    :param arr:
    :param band_set: dictionary with bands and points along the spectrum
    :param orientation: edc or mdc
    :param direction_normal:
    :param preferred_k_direction:
    :param dataset:
    :return: Dataset or DataArray, as controlled by the parameter "dataset"
    """

    if background:
        from arpes.models.band import AffineBackgroundBand
        background = AffineBackgroundBand

    free_directions = list(arr.dims)
    free_directions.remove(fit_direction)

    def is_between(x, y0, y1):
        y0, y1 = np.min([y0, y1]), np.max([y0, y1])
        return y0 <= x <= y1

    def interpolate_itersecting_fragments(coord, coord_index, points):
        """
        Finds all consecutive pairs of points in `points`
        :param coord:
        :param coord_idx:
        :param points:
        :return:
        """

        assert len(points[0]) == 2 # only support 2D interpolation

        for point_low, point_high in zip(points, points[1:]):
            coord_other_index = 1 - coord_index

            check_coord_low, check_coord_high = point_low[coord_index], point_high[coord_index]
            if is_between(coord, check_coord_low, check_coord_high):
                # this is unnecessarily complicated
                if check_coord_low < check_coord_high:
                    yield coord, (coord - check_coord_low) / (check_coord_high - check_coord_low) * \
                          (point_high[coord_other_index] - point_low[coord_other_index]) + \
                          point_low[coord_other_index]
                else:
                    yield coord, (coord - check_coord_high) / (check_coord_low - check_coord_high) * \
                          (point_low[coord_other_index] - point_high[coord_other_index]) + \
                          point_high[coord_other_index]


    def resolve_partial_bands_from_description(
            coord_dict, name=None, band=arpes.models.band.Band, dims=None,
            params=None, points=None, marginal=None):
        # You don't need to supply a marginal, but it is useful because it allows estimation of the initial value for
        # the amplitude from the approximate peak location

        if params is None:
            params = {}


        coord_name = [d for d in dims if d in coord_dict][0]
        iter_coord_value = coord_dict[coord_name]
        partial_band_locations = list(interpolate_itersecting_fragments(
            iter_coord_value, arr.dims.index(coord_name), points or []))

        def build_params(old_params, center, center_stray=None):
            new_params = copy.deepcopy(old_params)
            new_params.update({
                'center': {'value': center,}
            })
            if center_stray is not None:
                new_params['center']['min'] = center - center_stray
                new_params['center']['max'] = center + center_stray
                new_params['sigma'] = new_params.get('sigma', {})
                new_params['sigma']['value'] = center_stray
                if marginal is not None:
                    near_center = marginal.sel(**dict([[marginal.dims[0], slice(
                        center - 1.2 * center_stray,
                        center + 1.2 * center_stray,
                    )]]))

                    low, high = np.percentile(near_center.values, (20, 80,))
                    new_params['amplitude'] = new_params.get('amplitude', {})
                    new_params['amplitude']['value'] = high - low
            return new_params

        return [{
            'band': band,
            'name': '{}_{}'.format(name, i),
            'params': build_params(params, band_center, params.get('stray', stray)), # TODO
        } for i, (_, band_center) in enumerate(partial_band_locations)]

    template = arr.sum(fit_direction)
    band_results = xr.DataArray(
        np.ndarray(shape=template.values.shape, dtype=object),
        coords=template.coords,
        dims=template.dims,
        attrs=template.attrs
    )

    total_slices = np.product([len(arr.coords[d]) for d in free_directions])
    for coord_dict, marginal in wrap_tqdm(arr.T.iterate_axis(free_directions), interactive,
                                          desc='fitting',
                                          total=total_slices):
        partial_bands = [resolve_partial_bands_from_description(
            coord_dict, marginal=marginal, **b) for b in band_set.values()]

        partial_bands = [p for p in partial_bands if len(p)]

        if background is not None and partial_bands:
            partial_bands = partial_bands + [[{
                'band': background,
                'name': '',
                'params': {},
            }]]

        def instantiate_band(partial_band):
            phony_band = partial_band['band'](partial_band['name'])
            built = phony_band.fit_cls(prefix=partial_band['name'], missing='drop')
            for constraint_coord, params in partial_band['params'].items():
                if constraint_coord == 'stray':
                    continue
                built.set_param_hint(constraint_coord, **params)
            return built

        internal_models = [instantiate_band(b) for bs in partial_bands for b in bs]

        if not internal_models:
            band_results.loc[coord_dict] = None
            continue

        composite_model = functools.reduce(lambda x, y: x + y, internal_models)
        new_params = composite_model.make_params()
        fit_result = composite_model.fit(marginal.values, new_params,
                                         x=marginal.coords[list(marginal.indexes)[0]].values)

        # populate models, sample code
        band_results.loc[coord_dict] = fit_result

    if not dataset:
        band_results.attrs['original_data'] = arr
        return band_results

    residual = arr.copy(deep=True)
    residual.values = np.zeros(residual.shape)

    for coords in band_results.T.iter_coords():
        fit_item = band_results.sel(**coords).item()
        if fit_item is None:
            continue

        try:
            residual.loc[coords] = fit_item.residual
        except:
            pass

    return xr.Dataset({
        'data': arr,
        'residual': residual,
        'results': band_results,
        'norm_residual': residual / arr,
    }, residual.coords)


def fit_bands(arr: xr.DataArray, band_description, background=None,
              direction='mdc', preferred_k_direction=None, step=None):
    """
    Fits bands and determines dispersion in some region of a spectrum
    :param arr:
    :param band_description: A description of the bands to fit in the region
    :param background:
    :param direction:
    :return:
    """
    assert(direction in ['edc', 'mdc'])

    def iterate_marginals(arr: xr.DataArray, iterate_directions=None):
        if iterate_directions is None:
            iterate_directions = list(arr.dims)
            iterate_directions.remove('eV')

        selectors = itertools.product(*[arr.coords[d] for d in iterate_directions])
        for ss in selectors:
            coords = dict(zip(iterate_directions, [float(s) for s in ss]))
            yield arr.sel(**coords), coords

    directions = list(arr.dims)

    broadcast_direction = 'eV'

    if direction == 'mdc':
        if preferred_k_direction is None:
            possible_directions = set(directions).intersection({'kp', 'kx', 'ky', 'phi'})
            broadcast_direction = list(possible_directions)[0]

    directions.remove(broadcast_direction)

    residual, _ = next(iterate_marginals(arr, directions))
    residual = residual - np.min(residual.values)

    # Let the first band be given by fitting the raw data to this band
    # Find subsequent peaks by fitting models to the residuals
    raw_bands = [band.get('band') if isinstance(band, dict) else band for band in band_description]
    initial_fits = None
    all_fit_parameters = {}

    if step == 'initial':
        residual.plot()

    for band in band_description:
        if isinstance(band, dict):
            band_inst = band.get('band')
            params = band.get('params', {})
        else:
            band_inst = band
            params = None
        fit_model = band_inst.fit_cls(prefix=band_inst.label)
        initial_fit = fit_model.guess_fit(residual, params=params)
        if initial_fits is None:
            initial_fits = initial_fit.params
        else:
            initial_fits.update(initial_fit.params)

        residual = residual - initial_fit.best_fit
        if isinstance(band_inst, arpes.models.band.BackgroundBand):
            # This is an approximation to simulate a constant background band underneath the data
            # Because backgrounds are added to our model only after the initial sequence of fits.
            # This is by no means the most appropriate way to do this, just one that works
            # alright for now
            pass
            #residual = residual - residual.min()

        if step == 'initial':
            residual.plot()
            (residual - residual + initial_fit.best_fit).plot()

    if step == 'initial':
        return None, None, residual

    template = arr.sum(broadcast_direction)
    band_results = xr.DataArray(
        np.ndarray(shape=template.values.shape, dtype=object),
        coords=template.coords,
        dims=template.dims,
        attrs=template.attrs
    )

    for marginal, coordinate in iterate_marginals(arr, directions):
        # Use the closest parameters that have been successfully fit, or use the initial
        # parameters, this should be good enough because the order of the iterator will
        # be stable
        closest_model_params = initial_fits # fix me
        dist = float('inf')
        frozen_coordinate = tuple(coordinate[k] for k in template.dims)
        for c, v in all_fit_parameters.items():
            delta = np.array(c) - frozen_coordinate
            current_distance = delta.dot(delta)
            if current_distance < dist:
                # current_distance = dist
                if direction == 'mdc': # TODO remove me
                    closest_model_params = v

        closest_model_params = copy.deepcopy(closest_model_params)

        # TODO mix in any params to the model params

        # populate models
        internal_models = [band.fit_cls(prefix=band.label) for band in raw_bands]
        composite_model = functools.reduce(lambda x, y: x + y, internal_models)
        new_params = composite_model.make_params(**{k: v.value for k, v in closest_model_params.items()})
        fit_result = composite_model.fit(marginal.values, new_params,
                                         x=marginal.coords[list(marginal.indexes)[0]].values)

        # insert fit into the results, insert the parameters into the cache so that we have
        # fitting parameters for the next sequence
        band_results.loc[coordinate] = fit_result
        all_fit_parameters[frozen_coordinate] = fit_result.params


    # Unpack the band results
    #unpacked_bands = unpack_bands_from_fit(band_results)
    unpacked_bands = None
    residual = None

    return band_results, unpacked_bands, residual



import copy
import itertools
import os.path
import warnings

import numpy as np
from astropy.io import fits

import arpes.config
import xarray as xr
from arpes.endstations import (HemisphericalEndstation, SynchrotronEndstation,
                               find_clean_coords)
from arpes.provenance import provenance_from_file
from arpes.utilities import rename_keys

__all__ = ('HERSEndstation',)

class HERSEndstation(SynchrotronEndstation, HemisphericalEndstation):
    """
    This should be unified with the FITs endstation code, but I don't have any projects at BL10
    at the moment so I will defer the complexity of unifying them for now
    """

    PRINCIPAL_NAME = 'ALS-BL1001'
    ALIASES = ['ALS-BL1001', 'HERS', 'ALS-HERS', 'BL1001']

    def load(self, scan_desc: dict=None, **kwargs):
        if scan_desc is None:
            warnings.warn('Attempting to make due without user associated scan_desc for the file')
            raise TypeError('Expected a dictionary of scan_desc with the location of the file')

        scan_desc = dict(copy.deepcopy(scan_desc))

        data_loc = scan_desc.get('path', scan_desc.get('file'))
        data_loc = data_loc if data_loc.startswith('/') else os.path.join(arpes.config.DATA_PATH, data_loc)

        hdulist = fits.open(data_loc)

        hdulist[0].verify('fix+warn')
        header_hdu, hdu = hdulist[0], hdulist[1]

        coords, dimensions, spectrum_shape = find_clean_coords(hdu, scan_desc)
        columns = hdu.columns # pylint: disable=no-member

        column_renamings = {}
        take_columns = columns

        spectra_names = [name for name in take_columns if name in columns.names]

        skip_frags = {}
        skip_predicates = {lambda k: any(s in k for s in skip_frags)}
        scan_desc = {k: v for k, v in scan_desc.items()
                    if not any(pred(k) for pred in skip_predicates)}

        data_vars = {k: (dimensions[k], hdu.data[k].reshape(spectrum_shape[k]), scan_desc)  # pylint: disable=no-member
                     for k in spectra_names}
        data_vars = rename_keys(data_vars, column_renamings)

        hdulist.close()

        relevant_dimensions = {k for k in coords.keys() if k in
                               set(itertools.chain(*[l[0] for l in data_vars.values()]))}
        relevant_coords = {k: v for k, v in coords.items() if k in relevant_dimensions}

        deg_to_rad_coords = {'beta', 'psi', 'chi', 'theta'}
        relevant_coords = {k: c * (np.pi / 180) if k in deg_to_rad_coords else c
                           for k, c in relevant_coords.items()}

        dataset = xr.Dataset(
            data_vars,
            relevant_coords,
            scan_desc,
        )

        provenance_from_file(dataset, data_loc, {
            'what': 'Loaded BL10 dataset',
            'by': 'load_DLD',
        })

        return dataset



import numpy as np

from arpes.analysis.general import rebin
from arpes.analysis.savitzky_golay import savitzky_golay
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('approximate_core_levels',)


def local_minima(a, promenance=3):
    """
    Calculates local minima (maxima) according to a prominence criterion. The point should be lower
    than any in the region around it.

    Rather than searching manually, we perform some fancy indexing to do the calculation
    across the whole array simultaneously and iterating over the promenance criterion instead of
    through the data and the promenance criterion.
    :param a:
    :param promenance:
    :return:
    """
    conditions = a == a
    for i in range(1, promenance + 1):
        current_conditions = np.r_[[False] * i, a[i:] < a[:-i]] & np.r_[a[:-i] < a[i:], [False] * i]
        conditions = conditions & current_conditions

    return conditions


def local_maxima(a, promenance=3):
    return local_minima(-a, promenance)


local_maxima.__doc__ = local_minima.__doc__


def approximate_core_levels(data: DataType, window_size=None, order=5, binning=3, promenance=5):
    """
    Approximately locates core levels in a spectrum. Data is first smoothed, and then local
    maxima with sufficient prominence over other nearby points are selected as peaks.

    This can be helfpul to "seed" a curve fitting analysis for XPS.
    :param data:
    :param window_size:
    :param order:
    :param binning:
    :param promenance:
    :return:
    """
    data = normalize_to_spectrum(data)

    dos = data.S.sum_other(['eV']).sel(eV=slice(None, -20))

    if window_size is None:
        window_size = int(len(dos) / 40) # empirical, may change
        if window_size % 2 == 0:
            window_size += 1

    smoothed = rebin(savitzky_golay(dos, window_size, order), eV=binning)

    indices = np.argwhere(local_maxima(smoothed.values, promenance=promenance))
    energies = [smoothed.coords['eV'][idx].item() for idx in indices]

    return energies



import collections

import numpy as np

from typing import Any, Dict

__all__ = ('deep_equals', 'deep_update',)


def deep_update(destination: Any, source: Any) -> Dict[str, Any]:
    """
    Doesn't clobber keys further down trees like doing a shallow update would. Instead recurse down from the root
    and update as appropriate.
    :param destination:
    :param source:
    :return:
    """
    for k, v in source.items():
        if isinstance(v, collections.Mapping):
            destination[k] = deep_update(destination.get(k, {}), v)
        else:
            destination[k] = v

    return destination


def deep_equals(a: Any, b: Any) -> bool:
    if not isinstance(b, type(a)):
        print(b, a)
        return False

    if isinstance(a, (int, str, float, np.float, np.int, np.float64, np.int64,),):
        return a == b

    if a is None:
        return b is None

    if not isinstance(a, (dict, list, tuple, set,)):
        raise TypeError('Only dict, list, tuple, and set are supported by deep_equals, not {}'.format(type(a)))

    if isinstance(a, set):
        for item in a:
            if item not in b:
                return False

        for item in b:
            if item not in a:
                return False

        return True

    if isinstance(a, (list, tuple)):
        if len(a) != len(b):
            return False

        for i in range(len(a)):
            item_a, item_b = a[i], b[i]

            if not deep_equals(item_a, item_b):
                return False

        return True

    if isinstance(a, dict):
        if not set(a.keys()) == set(b.keys()):
            return False

        for k in a.keys():
            item_a, item_b = a[k], b[k]

            if not deep_equals(item_a, item_b):
                return False

        return True



import numpy as np

import arpes.constants
import xarray as xr

__all__ = ('calculate_kp_kz_bounds', 'calculate_kx_ky_bounds', 'calculate_kp_bounds')


def euler_to_kx(kinetic_energy, phi, beta, theta=0, slit_is_vertical=False):
    if slit_is_vertical:
        return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(beta) * np.cos(phi)
    else:
        return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(phi + theta)


def euler_to_ky(kinetic_energy, phi, beta, theta=0, slit_is_vertical=False):
    if slit_is_vertical:
        return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * (
            np.cos(theta) * np.sin(phi) + np.cos(beta) * np.cos(phi) * np.sin(theta)
        )
    else:
        return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * (
            np.cos(phi + theta) * np.sin(beta),
        )


def euler_to_kz(kinetic_energy, phi, beta, theta=0, inner_potential=10, slit_is_vertical=False):
    if slit_is_vertical:
        beta_term = -np.sin(theta) * np.sin(phi) + np.cos(theta) * np.cos(beta) * np.cos(phi)

    else:
        beta_term = np.cos(phi + theta) * np.cos(beta)

    return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy * beta_term ** 2 + inner_potential)


def spherical_to_kx(kinetic_energy: np.float, theta: np.float, phi: np.float) -> np.float:
    return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(theta) * np.cos(phi)


def spherical_to_ky(kinetic_energy, theta, phi):
    return arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(theta) * np.sin(phi)


def spherical_to_kz(kinetic_energy: np.float, theta: np.float, phi: np.float, inner_V: np.float) -> np.float:
    r"""
    K_INV_ANGSTROM encodes that k_z = \frac{\sqrt{2 * m * E_kin * \cos^2\theta + V_0}}{\hbar}
    :param kinetic_energy:
    :param theta:
    :param phi:
    :param inner_V:
    :return:
    """
    return arpes.constants.K_INV_ANGSTROM * np.sqrt(
        kinetic_energy * np.cos(theta) ** 2 + inner_V)


def calculate_kp_kz_bounds(arr: xr.DataArray):
    phi_offset = arr.S.phi_offset
    phi_min = np.min(arr.coords['phi'].values) - phi_offset
    phi_max = np.max(arr.coords['phi'].values) - phi_offset

    binding_energy_min, binding_energy_max = np.min(arr.coords['eV'].values), np.max(arr.coords['eV'].values)
    hv_min, hv_max = np.min(arr.coords['hv'].values), np.max(arr.coords['hv'].values)

    wf = arr.S.work_function
    kx_min = min(spherical_to_kx(hv_max - binding_energy_max - wf, phi_min, 0),
                 spherical_to_kx(hv_min - binding_energy_max - wf, phi_min, 0))
    kx_max = max(spherical_to_kx(hv_max - binding_energy_max - wf, phi_max, 0),
                 spherical_to_kx(hv_min - binding_energy_max - wf, phi_max, 0))

    angle_max = max(abs(phi_min), abs(phi_max))
    inner_V = arr.S.inner_potential
    kz_min = spherical_to_kz(hv_min + binding_energy_min - wf, angle_max, 0, inner_V)
    kz_max = spherical_to_kz(hv_max + binding_energy_max - wf, 0, 0, inner_V)

    return (
        (round(kx_min, 2), round(kx_max, 2)), # kp
        (round(kz_min, 2), round(kz_max, 2)), # kz
    )


def calculate_kp_bounds(arr: xr.DataArray):
    phi_coords = arr.coords['phi'].values - arr.S.phi_offset
    beta = float(arr.attrs.get('beta')) - arr.S.beta_offset

    phi_low, phi_high = np.min(phi_coords), np.max(phi_coords)
    phi_mid = (phi_high + phi_low) / 2

    sampled_phi_values = np.array([phi_low, phi_mid, phi_high])

    kinetic_energy = arr.S.hv - arr.S.work_function
    kps = arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(sampled_phi_values) * np.cos(beta)

    return round(np.min(kps), 2), round(np.max(kps), 2)


def calculate_kx_ky_bounds(arr: xr.DataArray):
    """
    Calculates the kx and ky range for a dataset with a fixed photon energy

    This is used to infer the gridding that should be used for a k-space conversion.
    Based on Jonathan Denlinger's old codes
    :param arr: Dataset that includes a key indicating the photon energy of the scan
    :return: ((kx_low, kx_high,), (ky_low, ky_high,))
    """
    phi_coords, beta_coords = arr.coords['phi'] - arr.S.phi_offset, arr.coords['beta'] - arr.S.beta_offset

    # Sample hopefully representatively along the edges
    phi_low, phi_high = np.min(phi_coords), np.max(phi_coords)
    beta_low, beta_high = np.min(beta_coords), np.max(beta_coords)
    phi_mid = (phi_high + phi_low) / 2
    beta_mid = (beta_high + beta_low) / 2

    sampled_phi_values = np.array([phi_high, phi_high, phi_mid, phi_low, phi_low,
                                   phi_low, phi_mid, phi_high, phi_high])
    sampled_beta_values = np.array([beta_mid, beta_high, beta_high, beta_high, beta_mid,
                                     beta_low, beta_low, beta_low, beta_mid])
    kinetic_energy = arr.S.hv - arr.S.work_function

    kxs = arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.sin(sampled_phi_values)
    kys = arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy) * np.cos(sampled_phi_values) * np.sin(sampled_beta_values)

    return (
        (round(np.min(kxs), 2), round(np.max(kxs), 2)),
        (round(np.min(kys), 2), round(np.max(kys), 2)),
    )



"""
.. module:: gap
   :synopsis: Utilities for gap fitting in ARPES, contains tools to normalize by Fermi-Dirac occupation

.. moduleauthor:: Conrad Stansbury <chstan@berkeley.edu>
"""
import warnings

import numpy as np

import xarray as xr
from arpes.constants import K_BOLTZMANN_EV_KELVIN
from arpes.fits.fit_models import AffineBroadenedFD
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('normalize_by_fermi_dirac', 'determine_broadened_fermi_distribution',
           'symmetrize')


def determine_broadened_fermi_distribution(reference_data: DataType, fixed_temperature=True):
    """
    Determine the parameters for broadening by temperature and instrumental resolution
    for a piece of data.

    As a general rule, we first try to estimate the instrumental broadening and linewidth broadening
    according to calibrations provided for the beamline + instrument, as a starting point.

    We also calculate the thermal broadening to expect, and fit an edge location. Then we use a Gaussian
    convolved Fermi-Dirac distribution against an affine density of states near the Fermi level, with a constant
    offset background above the Fermi level as a simple but effective model when away from lineshapes.

    These parameters can be used to bootstrap a fit to actual data or used directly in ``normalize_by_fermi_dirac``.


    :param reference_data:
    :return:
    """
    params = {}

    if fixed_temperature:
        params['fd_width'] = {
            'value': reference_data.S.temp * K_BOLTZMANN_EV_KELVIN,
            'vary': False,
        }

    reference_data = normalize_to_spectrum(reference_data)

    sum_dims = list(reference_data.dims)
    sum_dims.remove('eV')

    return AffineBroadenedFD().guess_fit(reference_data.sum(sum_dims), params=params)


@update_provenance('Normalize By Fermi Dirac')
def normalize_by_fermi_dirac(data: DataType, reference_data: DataType = None, plot=False,
                             broadening=None,
                             temperature_axis=None,
                             temp_offset=0, **kwargs):
    """
    Normalizes data according to a Fermi level reference on separate data or using the same source spectrum.

    To do this, a linear density of states is multiplied against a resolution broadened Fermi-Dirac
    distribution (`arpes.fits.fit_models.AffineBroadenedFD`). We then set the density of states to 1 and
    evaluate this model to obtain a reference that the desired spectrum is normalized by.

    :param data: Data to be normalized.
    :param reference_data: A reference spectrum, typically a metal reference. If not provided the
                           integrated data is used. Beware: this is inappropriate if your data is gapped.
    :param plot: A debug flag, allowing you to view the normalization spectrum and relevant curve-fits.
    :param broadening: Detector broadening.
    :param temperature_axis: Temperature coordinate, used to adjust the quality
                             of the reference for temperature dependent data.
    :param temp_offset: Temperature calibration in the case of low temperature data. Useful if the
                        temperature at the sample is known to be hotter than the value recorded off of a diode.
    :param kwargs:
    :return:
    """
    reference_data = data if reference_data is None else reference_data
    broadening_fit = determine_broadened_fermi_distribution(reference_data, **kwargs)
    broadening = broadening_fit.params['conv_width'].value if broadening is None else broadening

    if plot:
        print('Gaussian broadening is: {} meV (Gaussian sigma)'.format(
            broadening_fit.params['conv_width'].value * 1000))
        print('Fermi edge location is: {} meV (fit chemical potential)'.format(
            broadening_fit.params['fd_center'].value * 1000))
        print('Fermi width is: {} meV (fit fermi width)'.format(
            broadening_fit.params['fd_width'].value * 1000))

        broadening_fit.plot()

    offset = broadening_fit.params['offset'].value
    without_offset = broadening_fit.eval(offset=0)

    cut_index = -np.argmax(without_offset[::-1] > 0.1 * offset)
    cut_energy = reference_data.coords['eV'].values[cut_index]

    if temperature_axis is None and 'temp' in data.dims:
        temperature_axis = 'temp'

    transpose_order = list(data.dims)
    transpose_order.remove('eV')

    if temperature_axis:
        transpose_order.remove(temperature_axis)
        transpose_order = transpose_order + [temperature_axis]

    transpose_order = transpose_order + ['eV']

    without_background = (data - data.sel(eV=slice(cut_energy, None)).mean('eV')).transpose(*transpose_order)

    if temperature_axis:
        without_background = normalize_to_spectrum(without_background)
        divided = without_background.T.map_axes(
            temperature_axis, lambda x, coord: x / broadening_fit.eval(
                x=x.coords['eV'].values, lin_bkg=0, const_bkg=1, offset=0,
                conv_width=broadening,
                fd_width=(coord[temperature_axis] + temp_offset) * K_BOLTZMANN_EV_KELVIN))
    else:
        without_background = normalize_to_spectrum(without_background)
        divided = without_background / broadening_fit.eval(
            x=data.coords['eV'].values,
            conv_width=broadening,
            lin_bkg=0, const_bkg=1, offset=0)

    divided.coords['eV'].values = divided.coords['eV'].values - broadening_fit.params['fd_center'].value
    return divided


def _shift_energy_interpolate(data: DataType, shift=None):
    if shift is not None:
        pass
        # raise NotImplementedError("arbitrary shift not yet implemented")

    data = normalize_to_spectrum(data).S.transpose_to_front('eV')

    new_data = data.copy(deep=True)
    new_axis = new_data.coords['eV']
    new_values = new_data.values * 0

    if shift is None:
        closest_to_zero = data.coords['eV'].sel(eV=0, method='nearest')
        shift = -closest_to_zero

    stride = data.T.stride('eV', generic_dim_names=False)

    if np.abs(shift) >= stride:
        n_strides = int(shift / stride)
        new_axis = new_axis + n_strides * stride

        shift = shift - stride * n_strides

    new_axis = new_axis + shift

    weight = float(shift / stride)

    new_values = new_values + data.values * (1 - weight)
    if shift > 0:
        new_values[1:] = new_values[1:] + data.values[:-1] * weight
    if shift < 0:
        new_values[:-1] = new_values[:-1] + data.values[1:] * weight

    new_data.coords['eV'] = new_axis
    new_data.values = new_values

    return new_data


@update_provenance('Symmetrize')
def symmetrize(data: DataType, subpixel=False, full_spectrum=False):
    """
    Symmetrizes data across the chemical potential. This provides a crude tool by which
    gap analysis can be performed. In this implementation, subpixel accuracy is achieved by
    interpolating data.

    :param data: Input array.
    :param subpixel: Enable subpixel correction
    :param full_spectrum: Returns data above and below the chemical potential. By default, only
           the bound part of the spectrum (below the chemical potential) is returned, because
           the other half is identical.
    :return:
    """
    data = normalize_to_spectrum(data).S.transpose_to_front('eV')

    if subpixel or full_spectrum:
        data = _shift_energy_interpolate(data)

    above = data.sel(eV=slice(0, None))
    below = data.sel(eV=slice(None, 0)).copy(deep=True)

    l = len(above.coords['eV'])

    zeros = below.values * 0
    zeros[-l:] = above.values[::-1]

    below.values = below.values + zeros

    if full_spectrum:
        if not subpixel:
            warnings.warn("full spectrum symmetrization uses subpixel correction")

        full_data = below.copy(deep=True)

        new_above = full_data.copy(deep=True)[::-1]
        new_above.coords['eV'] = (new_above.coords['eV'] * -1)

        full_data = xr.concat([full_data, new_above[1:]], dim='eV')

        result = full_data
    else:
        result = below

    return result



import matplotlib.pyplot as plt

from arpes.provenance import save_plot_provenance

from .utils import path_for_plot, label_for_colorbar

__all__ = ('plot_with_bands',)

@save_plot_provenance
def plot_with_bands(data, bands, title=None, ax=None, norm=None,
                    out=None, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace('_', ' ')

    mesh = data.plot(norm=norm, ax=ax, **kwargs)
    mesh.colorbar.set_label(label_for_colorbar(data))

    if data.S.is_differentiated:
        mesh.set_cmap('Blues')

    for band in bands:
        plt.scatter(band.center.values, band.coords[band.dims[0]].values)

    if out is not None:
        filename = path_for_plot(out)
        plt.savefig(filename)
        return filename
    else:
        plt.show()
        return ax



import numpy as np

import arpes.constants
from typing import Any, Callable

from .base import CoordinateConverter, K_SPACE_BORDER, MOMENTUM_BREAKPOINTS
from .bounds_calculations import calculate_kp_kz_bounds

__all__ = ['ConvertKpKzV0', 'ConvertKxKyKz', 'ConvertKpKz']

class ConvertKpKzV0(CoordinateConverter):
    # TODO implement
    def __init__(self, *args, **kwargs):
        super(ConvertKpKzV0, self).__init__(*args, **kwargs)
        raise NotImplementedError()


class ConvertKxKyKz(CoordinateConverter):
    def __init__(self, *args, **kwargs):
        super(ConvertKxKyKz, self).__init__(*args, **kwargs)


class ConvertKpKz(CoordinateConverter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(ConvertKpKz, self).__init__(*args, **kwargs)
        self.hv = None

    def get_coordinates(self, resolution: dict=None):
        if resolution is None:
            resolution = {}

        coordinates = super(ConvertKpKz, self).get_coordinates(resolution)

        ((kp_low, kp_high), (kz_low, kz_high)) = calculate_kp_kz_bounds(self.arr)

        inferred_kp_res = (kp_high - kp_low + 2 * K_SPACE_BORDER) / len(self.arr.coords['phi'])
        inferred_kp_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kp_res][-1]

        # go a bit finer here because it would otherwise be very coarse
        inferred_kz_res = (kz_high - kz_low + 2 * K_SPACE_BORDER) / len(self.arr.coords['hv'])
        inferred_kz_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kz_res][-1]

        coordinates['kp'] = np.arange(kp_low - K_SPACE_BORDER, kp_high + K_SPACE_BORDER,
                                      resolution.get('kp', inferred_kp_res))
        coordinates['kz'] = np.arange(kz_low - K_SPACE_BORDER, kz_high + K_SPACE_BORDER,
                                      resolution.get('kz', inferred_kz_res))

        base_coords = {k: v for k, v in self.arr.coords.items()
                       if k not in ['eV', 'phi', 'hv']}

        coordinates.update(base_coords)

        return coordinates

    def kspace_to_hv(self, binding_energy: np.ndarray, kp: np.ndarray, kz: np.ndarray, *args: Any, **kwargs: Any) -> np.ndarray:
        # x = kp, y = kz, z = BE
        if self.hv is None:
            inner_v = self.arr.S.inner_potential
            wf = self.arr.S.work_function
            self.hv = arpes.constants.HV_CONVERSION * (kp ** 2 + kz ** 2) + (
                -inner_v - binding_energy + wf)

        return self.hv

    def kspace_to_phi(self, binding_energy: np.ndarray, kp: np.ndarray, kz: np.ndarray, *args: Any, **kwargs: Any) -> np.ndarray:
        if self.hv is None:
            self.kspace_to_hv(binding_energy, kp, kz, *args, **kwargs)

        def kp_to_polar(kinetic_energy_out: np.ndarray, kp: np.ndarray) -> np.ndarray:
            return np.arcsin(kp / (arpes.constants.K_INV_ANGSTROM * np.sqrt(kinetic_energy_out)))

        return kp_to_polar(self.hv + self.arr.S.work_function, kp) + self.arr.S.phi_offset

    def conversion_for(self, dim: str) -> Callable:
        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            'eV': self.kspace_to_BE,
            'hv': self.kspace_to_hv,
            'phi': self.kspace_to_phi,
        }.get(dim, with_identity)



"""
This module is a bit of a misnomer, in that it also applies perfectly well to data collected by a delay
line on a hemisphere, the important point is that the data in any given channel should correspond to the true number of
electrons that arrived in that channel.

Plotting routines here are ones that include statistical errorbars. Generally for datasets in PyARPES, an xr.Dataset
will hold the standard deviation data for a given variable on `{var_name}_std`.
"""
import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.plotting.utils import path_for_plot
from arpes.provenance import save_plot_provenance
from arpes.typing import DataType

__all__ = ('plot_with_std', 'scatter_with_std',)

@save_plot_provenance
def jitterplot(data: xr.Dataset, jitter_dimension=None):
    if jitter_dimension is None:
        jitter_dimension = 'bootstrap'

    assert jitter_dimension in data.coords



@save_plot_provenance
def plot_with_std(data: DataType, name_to_plot=None, ax=None, out=None, **kwargs):
    if name_to_plot is None:
        var_names = [k for k in data.data_vars.keys() if '_std' not in k]
        assert len(var_names) == 1
        name_to_plot = var_names[0]
        assert (name_to_plot + '_std') in data.data_vars.keys()

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.pop('figsize', (7, 5,)))

    data.data_vars[name_to_plot].plot(ax=ax, **kwargs)
    x, y = data.data_vars[name_to_plot].T.to_arrays()

    std = data.data_vars[name_to_plot + '_std'].values
    ax.fill_between(x, y - std, y + std, alpha=0.3, **kwargs)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    ax.set_xlim([np.min(x), np.max(x)])

    return fig, ax


@save_plot_provenance
def scatter_with_std(data: DataType, name_to_plot=None, ax=None, fmt='o', out=None, **kwargs):
    if name_to_plot is None:
        var_names = [k for k in data.data_vars.keys() if '_std' not in k]
        assert len(var_names) == 1
        name_to_plot = var_names[0]
        assert (name_to_plot + '_std') in data.data_vars.keys()

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.pop('figsize', (7, 5,)))

    x, y = data.data_vars[name_to_plot].T.to_arrays()

    std = data.data_vars[name_to_plot + '_std'].values
    ax.errorbar(x, y, yerr=std, fmt=fmt, markeredgecolor='black', **kwargs)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    ax.set_xlim([np.min(x), np.max(x)])

    return fig, ax



"""
This module contains monkey-patched versions of functions from xrft until improvements are made upstream.
We don't generally use this module too much anyway, and it is not a default import.
"""

# pylint: disable=import-error,no-member

# TODO note that there is a slight bug in the inverse Fourier transform here where the coords are sometimes rolled

import warnings
import xrft

import dask as dsar
import numpy as np
import xarray as xr

from pandas.api.types import is_timedelta64_dtype
from xrft.xrft import _hanning

from arpes.provenance import provenance

# Don't export the xrft additions we monkey patch them onto that module directly.
__all__ = ('fft_filter',)

def fft_filter(data: xr.DataArray, stops):
    """
    Applies a brick wall filter at region in ``stops`` in the Fourier transform of data. Use with care.
    :param data:
    :param stops:
    :return:
    """

    # This won't tolerate inverse inverse filtering ;)
    kdata = xrft.dft(data)

    # this will produce a fair amount of ringing, Scipy isn't very clear about how to butterworth filter in Nd
    for stop in stops:
        kstop = {'freq_' + k if 'freq_' not in k else k: v for k, v in stop.items()} # be nice
        kdata.loc[kstop] = 0

    kkdata = xrft.idft(kdata)
    kkdata.values = np.real(kkdata.values)
    kkdata.values = kkdata.values - np.min(kkdata.values) + np.mean(data.values)

    filtered_arr = xr.DataArray(
        kkdata,
        data.coords,
        data.dims,
        attrs=data.attrs.copy()
    )

    if 'id' in filtered_arr:
        del filtered_arr.attrs['id']

        provenance(filtered_arr, data, {
            'what': 'Apply a filter in frequency space by brick walling coordinate regions.',
            'by': 'fft_filter',
            'stops': stops,
        })

    return filtered_arr


def _spacing(da, dims):
    """
    Verify correct spacing and return the spacing for each axis
    :param da:
    :return:
    """
    delta_x = []
    for d in dims:
        coord = da[d]
        diff = np.diff(coord)
        if is_timedelta64_dtype(diff):
            # convert to seconds so we get hertz
            diff = diff.astype('timedelta64[s]').astype('f8')
        delta = diff[0]
        if not np.allclose(diff, diff[0]):
            raise ValueError("Can't take Fourier transform because"
                             "coodinate %s is not evenly spaced" % d)
        delta_x.append(delta)

    return delta_x


def rdft(da, dim=None, shift=True, remove_mean=True, window=False):
    """
    Perform real discrete Fourier transform of xarray data-array `da` along the
    specified dimensions.

    Parameters
    ----------
    da : `xarray.DataArray`
        The data to be transformed
    dim : list (optional)
        The dimensions along which to take the transformation. If `None`, all
        dimensions will be transformed.
    shift : bool (optional)
        Whether to shift the fft output.
    remove_mean : bool (optional)
        If `True`, the mean across the transform dimensions will be subtracted
        before calculating the Fourier transform.

    Returns
    -------
    daft : `xarray.DataArray`
        The output of the Fourier transformation, with appropriate dimensions.
    """
    if np.isnan(da.values).any():
        raise ValueError("Data cannot take Nans")

    if dim is None:
        dim = da.dims

    # the axes along which to take ffts
    axis_num = [da.get_axis_num(d) for d in dim]

    N = [da.shape[n] for n in axis_num]
    delta_x = _spacing(da, dim)

    # calculate frequencies from coordinates
    k = [ np.fft.rfftfreq(Nx, dx) for (Nx, dx) in zip(N, delta_x) ]

    if remove_mean:
        da = da - da.mean(dim=dim)

    if window:
        da = _hanning(da, N)

    # is this the best way to check for dask?
    data = da.data
    if hasattr(data, 'dask'):
        assert len(axis_num)==1
        f = dsar.fft.rfft(data, axis=axis_num[0])
    else:
        f = np.fft.rfftn(data, axes=axis_num)

    if shift:
        f = np.fft.fftshift(f, axes=axis_num)
        k = [np.fft.fftshift(l) for l in k]

    # set up new coordinates for dataarray
    prefix = 'freq_'
    k_names = [prefix + d for d in dim]
    k_coords = {key: val for (key,val) in zip(k_names, k)}

    newdims = list(da.dims)
    for anum, d in zip(axis_num, dim):
        newdims[anum] = prefix + d

    newcoords = {}
    for d in newdims:
        if d in da.coords:
            newcoords[d] = da.coords[d]
        else:
            newcoords[d] = k_coords[d]

    dk = [l[1] - l[0] for l in k]

    new_attrs = da.attrs.copy()
    new_attrs['N'] = N
    return xr.DataArray(f, dims=newdims, coords=newcoords, attrs=new_attrs)



def irdft(da, dim=None, shift=True, window=False):
    """
    Perform inverse real discrete Fourier transform of xarray data-array `da` along the
    specified dimensions.

    Parameters
    ----------
    da : `xarray.DataArray`
        The data to be transformed
    dim : list (optional)
        The dimensions along which to take the transformation. If `None`, all
        dimensions will be transformed.
    shift : bool (optional)
        Whether to shift the fft output.

    Returns
    -------
    daft : `xarray.DataArray`
        The output of the Fourier transformation, with appropriate dimensions.
    """
    if np.isnan(da.values).any():
        raise ValueError("Data cannot take NaNs")

    if dim is None:
        dim = da.dims

    # the axes along which to take ffts
    axis_num = [da.get_axis_num(d) for d in dim]

    N = [da.shape[n] for n in axis_num]
    delta_x = _spacing(da, dim)

    # calculate frequencies from coordinates
    k = [ np.fft.fftfreq(Nx, dx) for (Nx, dx) in zip(N, delta_x) ]

    if window:
        da = _hanning(da, N)

    # the hard work
    #f = np.fft.irfftn(da.values, axes=axis_num)
    # need special path for dask
    # is this the best way to check for dask?
    data = da.data
    if shift:
        data = np.fft.ifftshift(data, axes=axis_num)
        k = [np.fft.ifftshift(l) for l in k]

    if hasattr(data, 'dask'):
        assert len(axis_num)==1
        f = dsar.fft.irfft(data, axis=axis_num[0], s=da.attrs.get('N'))
    else:
        f = np.fft.irfftn(data, axes=axis_num, s=da.attrs.get('N'))

    # set up new coordinates for dataarray
    prefix = 'freq_'
    if any(prefix in d for d in da.dims):
        # came from a FFT we performed
        k_names = [d.split(prefix)[-1] for d in dim]
    else:
        k_names = [prefix + d for d in dim]

    k_coords = {key: val for (key,val) in zip(k_names, k)}

    newdims = list(da.dims)
    for anum, d in zip(axis_num, k_names):
        newdims[anum] = d

    newcoords = {}
    for d in newdims:
        if d in da.coords:
            newcoords[d] = da.coords[d]
        else:
            newcoords[d] = k_coords[d]

    dk = [l[1] - l[0] for l in k]

    new_attrs = da.attrs.copy()
    new_attrs['N'] = N
    return xr.DataArray(f, dims=newdims, coords=newcoords, attrs=new_attrs)


def idft(da, dim=None, shift=True, window=False):
    """
    Perform inverse real discrete Fourier transform of xarray data-array `da` along the
    specified dimensions.

    Parameters
    ----------
    da : `xarray.DataArray`
        The data to be transformed
    dim : list (optional)
        The dimensions along which to take the transformation. If `None`, all
        dimensions will be transformed.
    shift : bool (optional)
        Whether to shift the fft output.

    Returns
    -------
    daft : `xarray.DataArray`
        The output of the Fourier transformation, with appropriate dimensions.
    """
    warnings.warn('This is a little buggy, it does not produce an output with monotonic coords for some reason.')
    if np.isnan(da.values).any():
        raise ValueError("Data cannot take NaNs")

    if dim is None:
        dim = da.dims

    # the axes along which to take ffts
    axis_num = [da.get_axis_num(d) for d in dim]

    N = [da.shape[n] for n in axis_num]
    delta_x = _spacing(da, dim)

    # calculate frequencies from coordinates
    k = [ np.fft.fftfreq(Nx, dx) for (Nx, dx) in zip(N, delta_x) ]

    if window:
        da = _hanning(da, N)

    # the hard work
    #f = np.fft.ifftn(da.values, axes=axis_num)
    # need special path for dask
    # is this the best way to check for dask?
    data = da.data
    if shift:
        data = np.fft.ifftshift(data, axes=axis_num)
        k = [np.fft.ifftshift(l) for l in k]

    if hasattr(data, 'dask'):
        assert len(axis_num)==1
        f = dsar.fft.ifft(data, axis=axis_num[0], s=da.attrs.get('N'))
    else:
        f = np.fft.ifftn(data, axes=axis_num, s=da.attrs.get('N'))

    # set up new coordinates for dataarray
    prefix = 'freq_'
    if any(prefix in d for d in da.dims):
        # came from a FFT we performed
        k_names = [d.split(prefix)[-1] for d in dim]
    else:
        k_names = [prefix + d for d in dim]

    k_coords = {key: val for (key,val) in zip(k_names, k)}

    newdims = list(da.dims)
    for anum, d in zip(axis_num, k_names):
        newdims[anum] = d

    newcoords = {}
    for d in newdims:
        if d in da.coords:
            newcoords[d] = da.coords[d]
        else:
            newcoords[d] = k_coords[d]

    dk = [l[1] - l[0] for l in k]

    new_attrs = da.attrs.copy()
    new_attrs['N'] = N
    return xr.DataArray(f, dims=newdims, coords=newcoords, attrs=new_attrs)


# Monkey Patch
xrft.rdft = rdft
xrft.irdft = irdft
xrft.idft = idft



import numpy as np

from arpes.preparation import normalize_dim
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('find_t0', 'relative_change', 'normalized_relative_change')


@update_provenance('Normalized subtraction map')
def normalized_relative_change(data: DataType, t0=None, buffer=0.3, normalize_delay=True):
    """
    Calculates a normalized relative change, obtained by normalizing along the pump-probe "delay"
    axis and then subtracting the mean before t0 data and dividing by the original spectrum.
    :param data:
    :param t0:
    :param buffer:
    :param normalize_delay:
    :return:
    """
    spectrum = normalize_to_spectrum(data)
    if normalize_delay:
        spectrum = normalize_dim(spectrum, 'delay')
    subtracted = relative_change(spectrum, t0, buffer, normalize_delay=False)
    normalized = subtracted / spectrum
    normalized.values[np.isinf(normalized.values)] = 0
    normalized.values[np.isnan(normalized.values)] = 0
    return normalized


@update_provenance('Created simple subtraction map')
def relative_change(data: DataType, t0=None, buffer=0.3, normalize_delay=True):
    """
    Like normalized_relative_change, but only subtracts the before t0 data rather than
    normalizing by the original spectrum's intensity in each frame.
    :param data:
    :param t0:
    :param buffer:
    :param normalize_delay:
    :return:
    """
    spectrum = normalize_to_spectrum(data)
    if normalize_delay:
        spectrum = normalize_dim(spectrum, 'delay')

    delay_coords = spectrum.coords['delay']
    delay_start = np.min(delay_coords)

    if t0 is None:
        t0 = spectrum.S.t0 or find_t0(spectrum)

    assert t0 - buffer > delay_start

    before_t0 = spectrum.sel(delay=slice(None, t0 - buffer))
    subtracted = spectrum - before_t0.mean('delay')
    return subtracted


def find_t0(data: DataType, e_bound=0.02, approx=True):
    """
    Attempts to find the effective t0 in a spectrum by fitting a peak to the counts that occur
    far enough above e_F
    :param data:
    :param e_bound: Lower bound on the energy to use for the fitting
    :return:
    """
    spectrum = normalize_to_spectrum(data)

    assert 'delay' in spectrum.dims
    assert 'eV' in spectrum.dims

    sum_dims = set(spectrum.dims)
    sum_dims.remove('delay')
    sum_dims.remove('eV')

    summed = spectrum.sum(list(sum_dims)).sel(eV=slice(e_bound, None)).mean('eV')
    coord_max = summed.argmax().item()

    return summed.coords['delay'].values[coord_max]



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.tri import TriAnalyzer, Triangulation, UniformTriRefiner

from arpes.analysis import filters
from arpes.provenance import save_plot_provenance

from .utils import path_for_plot

__all__ = ['plot_isosurface', 'plot_trisurf', 'plotly_trisurf']


@save_plot_provenance
def plot_isosurface(data, level=None, percentile=99.5, smoothing=None, out=None, ax=None):
    assert 'eV' in data.dims

    new_dim_order = list(data.dims)
    new_dim_order.remove('eV')
    new_dim_order = new_dim_order + ['eV']
    data = data.transpose(*new_dim_order)
    colormap = plt.get_cmap('viridis')

    if smoothing is not None:
        data = filters.gaussian_filter_arr(data, sigma=smoothing)

    spacing = [data.coords[d][1] - data.coords[d][0] for d in data.dims]

    if level is None:
        level = np.percentile(data.data, percentile)

    # try to avoid skimage incompatibility with numpy v0.16 as much as possible
    from skimage import measure  # pylint: disable=import-error
    verts, faces, normals, values = measure.marching_cubes(data.values, level, spacing=spacing)

    if ax is None:
        fig = plt.figure(figsize=(7, 7))
        ax = fig.add_subplot(1, 1, 1, projection='3d')

    ax.plot_trisurf(verts[:, 0], verts[:,1], faces, verts[:, 2],
                    lw=1, facecolors=colormap(values))

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


def plotly_trisurf(data):
    import plotly.offline as pyo  # pylint: disable=import-error
    import plotly.graph_objs as go  # pylint: disable=import-error

    x = data.coords[data.dims[0]].values
    y = data.coords[data.dims[1]].values
    plot_data = [
        go.Surface(
            x=x,
            y=y,
            z=data.transpose().values,
        )
    ]
    layout = go.Layout(
        title='Test Title',
        autosize=True,
        width=900,
        height=900,
        margin=dict(
            l=65,
            r=50,
            b=65,
            t=90
        )
    )
    fig = go.Figure(data=plot_data, layout=layout)
    pyo.iplot(fig, filename='test-surface')


def plot_trisurf(data, out=None, ax=None, angles=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')

    Xs, Ys = np.meshgrid(data.coords[data.dims[0]].values, data.coords[data.dims[1]].values)
    triang = Triangulation(Xs.ravel(), Ys.ravel())
    mask = np.any(np.isnan(data.transpose().values.ravel()[triang.triangles]), axis=1)
    triang.set_mask(mask)

    values = data.copy(deep=True).transpose().values
    values[np.isnan(values)] = 0
    ax.plot_trisurf(triang, values.ravel(), antialiased=True, cmap=cm.coolwarm, linewidth=0)

    if angles is not None:
        ax.view_init(*angles)

    #plt.show()
    return ax

@save_plot_provenance
def plot_trisurface(data, out=None, ax=None):
    if ax is None:
        pass

    # -----------------------------------------------------------------------------
    # Generating the initial data test points and triangulation for the demo
    # -----------------------------------------------------------------------------
    # User parameters for data test points
    n_test = 200  # Number of test data points, tested from 3 to 5000 for subdiv=3

    subdiv = 3  # Number of recursive subdivisions of the initial mesh for smooth
    # plots. Values >3 might result in a very high number of triangles
    # for the refine mesh: new triangles numbering = (4**subdiv)*ntri

    init_mask_frac = 0.0  # Float > 0. adjusting the proportion of
    # (invalid) initial triangles which will be masked
    # out. Enter 0 for no mask.

    min_circle_ratio = .01  # Minimum circle ratio - border triangles with circle
    # ratio below this will be masked if they touch a
    # border. Suggested value 0.01 ; Use -1 to keep
    # all triangles.

    # Random points
    random_gen = np.random.mtrand.RandomState(seed=127260)
    x_test = random_gen.uniform(-1., 1., size=n_test)
    y_test = random_gen.uniform(-1., 1., size=n_test)
    z_test = experiment_res(x_test, y_test)

    # meshing with Delaunay triangulation
    x_dim, y_dim = tuple(data.dims)
    # x_coord, y_coord = data.coords[x_dim].data, data.coords[y_dim].data

    tri = Triangulation(data.coords[data.dims[0]].data, y_test)
    ntri = tri.triangles.shape[0]

    # Some invalid data are masked out
    mask_init = np.zeros(ntri, dtype=np.bool)
    masked_tri = random_gen.randint(0, ntri, int(ntri * init_mask_frac))
    mask_init[masked_tri] = True
    tri.set_mask(mask_init)

    # -----------------------------------------------------------------------------
    # Improving the triangulation before high-res plots: removing flat triangles
    # -----------------------------------------------------------------------------
    # masking badly shaped triangles at the border of the triangular mesh.
    mask = TriAnalyzer(tri).get_flat_tri_mask(min_circle_ratio)
    tri.set_mask(mask)

    # refining the data
    refiner = UniformTriRefiner(tri)
    tri_refi, z_test_refi = refiner.refine_field(z_test, subdiv=subdiv)

    # analytical 'results' for comparison
    z_expected = experiment_res(tri_refi.x, tri_refi.y)

    # for the demo: loading the 'flat' triangles for plot
    flat_tri = Triangulation(x_test, y_test)
    flat_tri.set_mask(~mask)

    # -----------------------------------------------------------------------------
    # Now the plots
    # -----------------------------------------------------------------------------
    # User options for plots
    plot_tri = True  # plot of base triangulation
    plot_masked_tri = True  # plot of excessively flat excluded triangles
    plot_refi_tri = False  # plot of refined triangulation
    plot_expected = False  # plot of analytical function values for comparison

    # Graphical options for tricontouring
    levels = np.arange(0., 1., 0.025)

    plt.figure()
    plt.gca().set_aspect('equal')
    plt.title("Filtering a Delaunay mesh\n" +
              "(application to high-resolution tricontouring)")

    # 1) plot of the refined (computed) data contours:
    plt.tricontour(tri_refi, z_test_refi, levels=levels, cmap='Blues',
                   linewidths=[2.0, 0.5, 1.0, 0.5])
    # 2) plot of the expected (analytical) data contours (dashed):
    if plot_expected:
        plt.tricontour(tri_refi, z_expected, levels=levels, cmap='Blues',
                       linestyles='--')
    # 3) plot of the fine mesh on which interpolation was done:
    if plot_refi_tri:
        plt.triplot(tri_refi, color='0.97')
    # 4) plot of the initial 'coarse' mesh:
    if plot_tri:
        plt.triplot(tri, color='0.7')
    # 4) plot of the unvalidated triangles from naive Delaunay Triangulation:
    if plot_masked_tri:
        plt.triplot(flat_tri, color='red')

    plt.show()

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()



from copy import deepcopy

import numpy as np

__all__ = ['remap_coords_to']


def remap_coords_to(arr, reference_arr):
    """
    This needs to be thought out a bit more, namely to take into account better the
    manipulator scan degree of freeedom.

    Produces coords which provide the scan cut path for the array ``arr`` as seen in the coordinate system defined
    by the manipulator location in ``reference_arr``. This is useful for plotting locations of cuts in a FS.

    This code also assumes that a hemispherical analyzer was used, because it uses the coordinate 'phi'.
    :param arr: Scan that represents a cut we would like to understand
    :param reference_arr: Scan providing the desired destination coordinates
    :return: Coordinates dict providing the path cut by the dataset ``arr``
    """

    irrelevant_coordinates = list({'hv', 'eV',}.intersection(set(arr.dims)))
    arr = arr.sum(*irrelevant_coordinates, keep_attrs=True) # sum is not so fast, but ensures there is data

    assert arr.S.is_kspace == reference_arr.S.is_kspace

    full_coords = arr.S.full_coords
    full_reference_coords = reference_arr.S.full_coords

    def float_or_zero(value):
        if isinstance(value, float):
            return value
        return 0

    delta_chi = float_or_zero(full_coords['chi']) - float_or_zero(full_reference_coords['chi'])
    delta_theta = float_or_zero(full_reference_coords['theta']) - float_or_zero(full_coords['theta'])

    if arr.S.is_kspace:
        # kspace
        raise NotImplementedError()
    else:
        # rotation matrix is
        # cos -sin
        # sin  cos
        o_phi = full_coords['phi'] + delta_theta
        o_polar = full_coords['beta']
        phi_coord = np.cos(-delta_chi) * o_phi - np.sin(-delta_chi) * o_polar
        polar_coord = np.sin(-delta_chi) * o_phi + np.cos(-delta_chi) * o_polar
        remapped_coords = deepcopy(full_reference_coords)
        remapped_coords.update({'phi': phi_coord.data, 'beta': polar_coord.data})
        return remapped_coords



from pprint import pprint

import numpy as np
import pandas as pd

from arpes.typing import DataType

__all__ = ('diff_attrs',)


def diff_attrs(a: DataType, b: DataType, should_print=True, skip_nan=False, skip_composite=True):
    attrs_a = a.attrs
    attrs_b = b.attrs

    a_has = {k: v for k, v in attrs_a.items() if k not in attrs_b}
    b_has = {k: v for k, v in attrs_b.items() if k not in attrs_a}

    def should_skip(k):
        if skip_composite:
            composites = (dict, list, np.ndarray, pd.DataFrame,)
            if isinstance(attrs_a[k], composites) or isinstance(attrs_b[k], composites):
                if type(attrs_a[k]) == type(attrs_b[k]):
                    return True

        try:
            if attrs_a[k] == attrs_b[k]:
                return True
        except ValueError:
            # probably a data frame
            return True

        if skip_nan and (np.isnan(attrs_a[k]) or np.isnan(attrs_b[k])):
            return True

        try:
            if np.isnan(attrs_a[k]) and np.isnan(attrs_b[k]):
                return True
        except:
            pass

        return False

    common = list(k for k in attrs_a.keys() if k in attrs_b and not should_skip(k))

    values_in_a = [attrs_a[k] for k in common]
    values_in_b = [attrs_b[k] for k in common]

    diff = pd.DataFrame(data={
        'key': common,
        'A': values_in_a,
        'B': values_in_b,
    }).set_index('key')


    if should_print:
        print('A has:')
        pprint(a_has)

        print('\nB has:')
        pprint(b_has)

        print('\nDifferences:')
        print(diff.to_string())
    else:
        return a_has, b_has, diff
